# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import datetime
import os
import pathlib
import re
import subprocess
import tempfile
import time
from typing import Any, Final, List, Mapping, Optional, Protocol, Union

from protenix.data.constants import STANDARD_POLYMER_CHAIN_TYPES
from protenix.data.tools.common import (
    convert_a3m_to_stockholm,
    convert_stockholm_to_a3m,
    lazy_fasta_parse,
)
from protenix.utils.logger import get_logger

logger = get_logger(__name__)

SHORT_SEQ_LIMIT: Final[int] = 50


def run_shell(
    cmd: List[str],
    name: str,
    **kwargs: Any,
) -> subprocess.CompletedProcess:
    """
    Run a shell command and log its execution and results.

    Args:
        cmd: List of command arguments.
        name: Logical name for the command (for logging).
        **kwargs: Additional keyword arguments for subprocess.run.

    Returns:
        CompletedProcess object.

    Raises:
        RuntimeError: If the command fails.
    """
    logger.info(f"Executing {name}: {' '.join(cmd)}")
    t0 = time.time()
    try:
        proc = subprocess.run(cmd, capture_output=True, text=True, check=True, **kwargs)
        logger.info(f"{name} finished in {time.time()-t0:.3f}s")
        return proc
    except subprocess.CalledProcessError as e:
        logger.error(f"{name} failed. Stderr: {e.stderr}")
        raise RuntimeError(
            f"{name} failed\nStdout: {e.stdout}\nStderr: {e.stderr}"
        ) from e


class BinaryWrapper:
    """
    Base class for wrapping external bioinformatics binaries.

    Args:
        binary_path: Path to the executable.
        name: Name of the tool.

    Raises:
        RuntimeError: If the binary path does not exist.
    """

    def __init__(self, binary_path: str, name: str) -> None:
        if not os.path.exists(binary_path):
            raise RuntimeError(f"{name} not found at {binary_path}")
        self.path = binary_path


class Hmmalign(BinaryWrapper):
    """
    Wrapper for HMMER hmmalign.

    Args:
        binary_path: Path to the hmmalign executable.
    """

    def __init__(self, binary_path: str) -> None:
        super().__init__(binary_path, "hmmalign")

    def align(
        self, a3m: str, profile: str, flags: Optional[Mapping[str, str]] = None
    ) -> str:
        """
        Align sequences in A3M format to a HMM profile.

        Args:
            a3m: Input sequences in A3M format.
            profile: HMM profile string.
            flags: Optional additional command line flags.

        Returns:
            Aligned sequences in A3M format.
        """
        with tempfile.TemporaryDirectory() as tmp:
            prof_p, in_p, out_p = f"{tmp}/p", f"{tmp}/i", f"{tmp}/o"
            pathlib.Path(prof_p).write_text(profile)
            pathlib.Path(in_p).write_text(a3m)
            cmd = [self.path, "-o", out_p, "--outformat", "A2M"]
            if flags:
                for k, v in flags.items():
                    cmd.extend([k, v])
            run_shell(cmd + [prof_p, in_p], "hmmalign")
            return pathlib.Path(out_p).read_text()

    def align_sequences_to_profile(self, profile: str, a3m: str) -> str:
        """
        Clean gaps and realign sequences to a profile.

        Args:
            profile: HMM profile string.
            a3m: Input sequences in A3M format.

        Returns:
            Realigned sequences.
        """
        clean_a3m = "\n".join(
            f">{d}\n{s.replace('-', '')}" for s, d in lazy_fasta_parse(a3m)
        )
        return self.align(clean_a3m, profile)


class Hmmbuild(BinaryWrapper):
    """
    Wrapper for HMMER hmmbuild.

    Args:
        binary_path: Path to the hmmbuild executable.
        singlemx: Whether to use --singlemx flag.
        alphabet: Optional alphabet specification (e.g., 'amino', 'rna', 'dna').
    """

    def __init__(
        self, binary_path: str, singlemx: bool = False, alphabet: Optional[str] = None
    ) -> None:
        super().__init__(binary_path, "hmmbuild")
        self.opts = (["--singlemx"] if singlemx else []) + (
            [f"--{alphabet}"] if alphabet else []
        )

    def build(self, msa: str, informat: str, model: str = "fast") -> str:
        """
        Build a HMM profile from a Multiple Sequence Alignment.

        Args:
            msa: Input MSA string.
            informat: Format of the input MSA (e.g., 'afa', 'stockholm').
            model: Model construction strategy ('fast' or 'hand').

        Returns:
            The HMM profile string.
        """
        with tempfile.TemporaryDirectory() as tmp:
            in_p, out_p = f"{tmp}/i", f"{tmp}/o"
            pathlib.Path(in_p).write_text(msa)
            cmd = [self.path, "--informat", informat] + self.opts
            if model == "hand":
                cmd.append("--hand")
            run_shell(cmd + [out_p, in_p], "hmmbuild")
            return pathlib.Path(out_p).read_text()

    def build_profile_from_a3m(self, a3m: str) -> str:
        """
        Build HMM from A3M format by removing insertions.

        Args:
            a3m: Input A3M string.

        Returns:
            HMM profile string.
        """
        clean = "".join(
            f">{d}\n{re.sub('[a-z]+', '', s)}\n" for s, d in lazy_fasta_parse(a3m)
        )
        return self.build(clean, "afa")

    def build_profile_from_sto(self, sto: str, model: str = "fast") -> str:
        """
        Build HMM from Stockholm format.

        Args:
            sto: Input Stockholm string.
            model: Model construction strategy.

        Returns:
            HMM profile string.
        """
        return self.build(sto, "stockholm", model)


class Hmmsearch(BinaryWrapper):
    """
    Wrapper for HMMER hmmsearch.

    Args:
        binary_path: Path to hmmsearch executable.
        hmmbuild_binary_path: Path to hmmbuild executable.
        database_path: Path to the sequence database to search.
        alphabet: Alphabet type.
        **filters: Search filters (e.g., e_value, inc_e, filter_f1, etc.).
    """

    def __init__(
        self,
        binary_path: str,
        hmmbuild_binary_path: str,
        database_path: str,
        alphabet: str = "amino",
        **filters: Any,
    ) -> None:
        super().__init__(binary_path, "hmmsearch")
        self.builder = Hmmbuild(binary_path=hmmbuild_binary_path, alphabet=alphabet)
        self.db = database_path
        self.flags = ["--max"] if filters.get("filter_max") else []
        if not self.flags:
            map_f = {
                "filter_f1": "--F1",
                "filter_f2": "--F2",
                "filter_f3": "--F3",
                "e_value": "-E",
                "inc_e": "--incE",
                "dom_e": "--domE",
                "incdom_e": "--incdomE",
            }
            for k, v in filters.items():
                if k in map_f and v is not None:
                    self.flags.extend([map_f[k], str(v)])

    def query_with_hmm(self, hmm: str) -> str:
        """
        Search a database using an HMM profile.

        Args:
            hmm: HMM profile string.

        Returns:
            Search result in A3M format.
        """
        with tempfile.TemporaryDirectory() as tmp:
            hmm_p, sto_p = f"{tmp}/q.hmm", f"{tmp}/o.sto"
            pathlib.Path(hmm_p).write_text(hmm)
            cmd = (
                [self.path, "--noali", "--cpu", "8"]
                + self.flags
                + ["-A", sto_p, hmm_p, self.db]
            )
            run_shell(cmd, "hmmsearch")
            with open(sto_p) as f:
                return convert_stockholm_to_a3m(f, remove_gaps=False, linewidth=60)

    def query_with_sto(self, sto: str, model: str = "fast") -> str:
        """
        Build an HMM from a Stockholm MSA and search the database.

        Args:
            sto: Input Stockholm MSA string.
            model: Model construction strategy.

        Returns:
            Search result in A3M format.
        """
        return self.query_with_hmm(self.builder.build_profile_from_sto(sto, model))


@dataclasses.dataclass(frozen=True, slots=True)
class MsaToolResult:
    """Data class for MSA tool results."""

    target_sequence: str
    e_value: Optional[float]
    a3m: str


class MsaTool(Protocol):
    """Protocol for MSA search tools."""

    def query(self, target_sequence: str) -> MsaToolResult:
        """Perform a sequence search query."""
        ...


class Jackhmmer(BinaryWrapper, MsaTool):
    """
    Wrapper for HMMER jackhmmer.

    Args:
        binary_path: Path to jackhmmer executable.
        database_path: Path to search database.
        n_cpu: Number of CPUs to use.
        n_iter: Number of iterations.
        e_value: E-value threshold.
        **kwargs: Additional parameters (e.g., z_value, filter_f1, etc.).
    """

    def __init__(
        self,
        binary_path: str,
        database_path: str,
        n_cpu: int = 8,
        n_iter: int = 3,
        e_value: Optional[float] = 1e-3,
        **kwargs: Any,
    ) -> None:
        super().__init__(binary_path, "jackhmmer")
        self.db, self.cpu, self.iter, self.e = database_path, n_cpu, n_iter, e_value
        self.extra = kwargs

    def query(self, seq: str) -> MsaToolResult:
        """
        Query a database using jackhmmer for a given sequence.

        Args:
            seq: Input sequence.

        Returns:
            MsaToolResult object.
        """
        with tempfile.TemporaryDirectory() as tmp:
            fa_p, sto_p = f"{tmp}/q.fa", f"{tmp}/o.sto"
            pathlib.Path(fa_p).write_text(f">query\n{seq}\n")
            cmd = [
                self.path,
                "-o",
                "/dev/null",
                "-A",
                sto_p,
                "--noali",
                "--cpu",
                str(self.cpu),
                "-N",
                str(self.iter),
            ]
            if self.e is not None:
                cmd.extend(["-E", str(self.e), "--incE", str(self.e)])
            for k, v in self.extra.items():
                flag = {
                    "filter_f1": "--F1",
                    "filter_f2": "--F2",
                    "filter_f3": "--F3",
                    "z_value": "-Z",
                }.get(k)
                if flag and v is not None:
                    cmd.extend([flag, str(v)])
            run_shell(cmd + [fa_p, self.db], "jackhmmer")
            with open(sto_p) as f:
                a3m = convert_stockholm_to_a3m(
                    f, max_seqs=self.extra.get("max_sequences", 5000)
                )
            return MsaToolResult(target_sequence=seq, e_value=self.e, a3m=a3m)


class Nhmmer(BinaryWrapper, MsaTool):
    """
    Wrapper for HMMER nhmmer.

    Args:
        binary_path: Path to nhmmer executable.
        hmmalign_binary_path: Path to hmmalign executable.
        hmmbuild_binary_path: Path to hmmbuild executable.
        database_path: Path to search database.
        n_cpu: Number of CPUs to use.
        e_value: E-value threshold.
        **kwargs: Additional parameters (e.g., alphabet, strand, filter_f3, etc.).
    """

    def __init__(
        self,
        binary_path: str,
        hmmalign_binary_path: str,
        hmmbuild_binary_path: str,
        database_path: str,
        n_cpu: int = 8,
        e_value: float = 1e-3,
        **kwargs: Any,
    ) -> None:
        super().__init__(binary_path, "nhmmer")
        self.aligner_p, self.builder_p, self.db, self.cpu, self.e = (
            hmmalign_binary_path,
            hmmbuild_binary_path,
            database_path,
            n_cpu,
            e_value,
        )
        self.opts = kwargs

    def query(self, seq: str) -> MsaToolResult:
        """
        Query a database using nhmmer for a given sequence.

        Args:
            seq: Input sequence.

        Returns:
            MsaToolResult object.
        """
        with tempfile.TemporaryDirectory() as tmp:
            fa_p, sto_p = f"{tmp}/q.fa", f"{tmp}/o.sto"
            pathlib.Path(fa_p).write_text(f">query\n{seq}\n")
            cmd = [
                self.path,
                "-o",
                "/dev/null",
                "--noali",
                "--cpu",
                str(self.cpu),
                "-E",
                str(self.e),
            ]
            if self.opts.get("alphabet"):
                cmd.append(f"--{self.opts['alphabet']}")
            if self.opts.get("strand"):
                cmd.append(f"--{self.opts['strand']}")
            f3 = (
                0.02
                if self.opts.get("alphabet") == "rna" and len(seq) < SHORT_SEQ_LIMIT
                else self.opts.get("filter_f3", 1e-5)
            )
            cmd.extend(["-A", sto_p, "--F3", str(f3), fa_p, self.db])
            run_shell(cmd, "nhmmer")

            if os.path.getsize(sto_p) > 0:
                with open(sto_p) as f:
                    a3m = convert_stockholm_to_a3m(
                        f, max_seqs=self.opts.get("max_sequences", 5000) - 1
                    )
                prof = Hmmbuild(
                    binary_path=self.builder_p, alphabet=self.opts.get("alphabet")
                ).build_profile_from_a3m(f">query\n{seq}\n")
                aligned = Hmmalign(
                    binary_path=self.aligner_p
                ).align_sequences_to_profile(prof, a3m)
                final_a3m = "\n".join(
                    f">{d}\n{s}"
                    for s, d in lazy_fasta_parse(f">query\n{seq}\n" + aligned)
                )
            else:
                final_a3m = f">query\n{seq}"
        return MsaToolResult(target_sequence=seq, e_value=self.e, a3m=final_a3m)


# Configuration Dataclasses
@dataclasses.dataclass(frozen=True, slots=True)
class DatabaseConfig:
    """Config for a biological sequence database."""

    name: str
    path: str


@dataclasses.dataclass(frozen=True, slots=True)
class JackhmmerConfig:
    """Config for a jackhmmer search run."""

    binary_path: str
    database_config: DatabaseConfig
    n_cpu: int
    n_iter: int
    e_value: float
    z_value: Optional[Union[float, int]]
    max_sequences: int


@dataclasses.dataclass(frozen=True, slots=True)
class NhmmerConfig:
    """Config for a nhmmer search run."""

    binary_path: str
    hmmalign_binary_path: str
    hmmbuild_binary_path: str
    database_config: DatabaseConfig
    n_cpu: int
    e_value: float
    max_sequences: int
    alphabet: Optional[str]


@dataclasses.dataclass(frozen=True, slots=True)
class RunConfig:
    """Generic config for an MSA search run."""

    config: Union[JackhmmerConfig, NhmmerConfig]
    chain_poly_type: str
    crop_size: Optional[int]

    def __post_init__(self) -> None:
        """Validate run configuration."""
        if self.crop_size is not None and self.crop_size < 2:
            raise ValueError("crop_size < 2")
        if self.chain_poly_type not in STANDARD_POLYMER_CHAIN_TYPES:
            raise ValueError("Invalid chain type")


@dataclasses.dataclass(frozen=True, slots=True)
class HmmsearchConfig:
    """Config for an hmmsearch run."""

    hmmsearch_binary_path: str
    hmmbuild_binary_path: str
    e_value: float = 100
    inc_e: float = 100
    dom_e: float = 100
    incdom_e: float = 100
    alphabet: str = "amino"
    filter_f1: Optional[float] = 0.1
    filter_f2: Optional[float] = 0.1
    filter_f3: Optional[float] = 0.1
    filter_max: bool = False


def run_hmmsearch_with_a3m(
    database_path: Union[str, os.PathLike],
    hmmsearch_config: HmmsearchConfig,
    max_a3m_query_sequences: Optional[int],
    a3m: Optional[str],
) -> str:
    """
    Search database using hmmsearch with an input A3M string.

    Args:
        database_path: Path to search database.
        hmmsearch_config: HmmsearchConfig object.
        max_a3m_query_sequences: Max sequences to use for building the HMM.
        a3m: Input A3M format string.

    Returns:
        Search results in A3M format.
    """
    cfg = dataclasses.asdict(hmmsearch_config)
    search_bin = cfg.pop("hmmsearch_binary_path")
    build_bin = cfg.pop("hmmbuild_binary_path")
    searcher = Hmmsearch(
        binary_path=search_bin,
        hmmbuild_binary_path=build_bin,
        database_path=database_path,
        **cfg,
    )
    sto = convert_a3m_to_stockholm(a3m, max_a3m_query_sequences)
    return searcher.query_with_sto(sto, model="hand")


# Legacy compatibility / Other classes if needed
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class TemplateToolConfig:
    """Configuration for template search tool."""

    database_path: str
    chain_poly_type: str
    hmmsearch_config: HmmsearchConfig
    max_a3m_query_sequences: Optional[int] = 300

    def __post_init__(self) -> None:
        """Validate template tool configuration."""
        if self.chain_poly_type not in STANDARD_POLYMER_CHAIN_TYPES:
            raise ValueError("Invalid chain type")


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class TemplateFilterConfig:
    """Configuration for filtering template search results."""

    max_subsequence_ratio: Optional[float]
    min_align_ratio: Optional[float]
    min_hit_length: Optional[int]
    deduplicate_sequences: bool
    max_hits: Optional[int]
    max_template_date: datetime.date

    @classmethod
    def no_op_filter(cls) -> "TemplateFilterConfig":
        """Return a filter config that doesn't filter anything."""
        return cls(
            max_subsequence_ratio=None,
            min_align_ratio=None,
            min_hit_length=None,
            deduplicate_sequences=False,
            max_hits=None,
            max_template_date=datetime.date(3000, 1, 1),
        )


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class TemplatesConfig:
    """Complete configuration for the template search pipeline."""

    template_tool_config: TemplateToolConfig
    filter_config: TemplateFilterConfig



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import torch

from protenix.model.modules.primitives import (
    rearrange_qk_to_dense_trunk,
    rearrange_to_dense_trunk,
)


def create_qkv(batch_size_dims, n_q, n_kv, d):
    q = torch.rand(size=(*batch_size_dims, n_q, d))
    k = torch.rand(size=(*batch_size_dims, n_kv, d))
    v = torch.rand(size=(*batch_size_dims, n_kv, d))
    return q, k, v


class TestUtils(unittest.TestCase):
    def setUp(self):
        return super().setUp()

    def test_equivalence(self):

        batch_size_dims = (3, 5)
        n = 128 * 2 + 18
        d = 9
        n_queries = 32
        n_keys = 128
        inf = 10e10

        torch.random.manual_seed(42)
        q, k, v = create_qkv(batch_size_dims, n, n, d)
        (
            q_trunked,
            k_trunked,
            _,
            attn_bias_trunked,
            q_pad_length,
        ) = rearrange_to_dense_trunk(
            q,
            k,
            v,
            n_queries,
            n_keys,
            inf=inf,
        )

        q_b, k_b, padding_info = rearrange_qk_to_dense_trunk(
            q, k, dim_q=-2, dim_k=-2, n_queries=n_queries, n_keys=n_keys
        )
        self.assertTrue(
            torch.allclose(
                padding_info["mask_trunked"] > 0, attn_bias_trunked[0, 0] > -1
            )
        )
        self.assertTrue(torch.allclose(q_b, q_trunked))
        self.assertTrue(torch.allclose(k_b, k_trunked))


if __name__ == "__main__":
    unittest.main()



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Any, Optional

import torch
import triton

from protenix.model.tri_attention.backward import (
    _attention_bwd_dbias2,
    _attention_bwd_dkdv,
    _attention_bwd_dq,
    _attention_bwd_preprocess,
)
from protenix.model.tri_attention.forward import _attention_fwd


def get_tag(x: int) -> int:
    if x > 4096:
        return 4096
    return math.ceil(x / 32) * 32


class TriAttentionFunction(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: Any,
        Q: torch.Tensor,
        K: torch.Tensor,
        V: torch.Tensor,
        Bias1: torch.Tensor,
        Bias2: torch.Tensor,
        deterministic: bool = False,
    ) -> torch.Tensor:
        Q = Q.contiguous()
        K = K.contiguous()
        V = V.contiguous()
        Bias1 = Bias1.contiguous()
        Bias2 = Bias2.contiguous()

        B, N, S, H, D = Q.shape
        TAG_N = get_tag(S)
        O = torch.empty_like(Q)
        M = torch.empty((B, N, S, H), device=Q.device, dtype=torch.float32)

        grid = lambda args: (triton.cdiv(S, args["BLOCK_M"]), H, B * N)
        _attention_fwd[grid](
            Q,
            K,
            V,
            O,
            Bias1,
            Bias2,
            M,
            N,
            S,
            H,
            HEAD_DIM=D,
            TAG_N=TAG_N,
            deterministic=deterministic,
        )

        ctx.save_for_backward(Q, K, V, Bias1, Bias2, O, M)
        ctx.deterministic = deterministic
        return O

    @staticmethod
    def backward(
        ctx: Any, DO: torch.Tensor
    ) -> tuple[
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        Optional[torch.Tensor],
        None,
    ]:
        Q, K, V, Bias1, Bias2, O, M = ctx.saved_tensors
        deterministic = ctx.deterministic
        B, N, S, H, D = Q.shape
        TAG_N = get_tag(S)
        DQ = torch.empty_like(Q)
        DK = torch.empty_like(K)
        DV = torch.empty_like(V)
        DBias1 = torch.empty_like(Bias1)
        DBias2 = torch.empty_like(Bias2)
        Delta = torch.empty_like(M)
        grid = lambda args: (triton.cdiv(S, args["BLOCK_M"]), H, B * N)
        _attention_bwd_preprocess[grid](O, DO, Delta, N, S, H, HEAD_DIM=D, TAG_N=TAG_N)

        grid = lambda args: (triton.cdiv(S, args["BLOCK_M"]), H, B * N)
        _attention_bwd_dkdv[grid](
            Q,
            K,
            V,
            Bias1,
            Bias2,
            M,
            Delta,
            DK,
            DV,
            DO,
            N,
            S,
            H,
            HEAD_DIM=D,
            TAG_N=TAG_N,
            deterministic=deterministic,
        )

        grid = lambda args: (triton.cdiv(S, args["BLOCK_M"]), H, B * N)
        _attention_bwd_dq[grid](
            Q,
            K,
            V,
            Bias1,
            Bias2,
            M,
            Delta,
            DQ,
            DO,
            N,
            S,
            H,
            HEAD_DIM=D,
            TAG_N=TAG_N,
            deterministic=deterministic,
        )

        grid = lambda args: (
            triton.cdiv(S, args["BLOCK_M"]),
            triton.cdiv(S, args["BLOCK_N"]),
            B * H,
        )
        _attention_bwd_dbias2[grid](
            Q,
            K,
            V,
            Bias1,
            Bias2,
            M,
            Delta,
            DBias2,
            DO,
            N,
            S,
            H,
            HEAD_DIM=D,
            TAG_N=TAG_N,
            deterministic=deterministic,
        )
        return DQ, DK, DV, DBias1, DBias2, None


class TriAttention(torch.nn.Module):
    def __init__(self, deterministic: bool = False) -> None:
        super(TriAttention, self).__init__()
        self.deterministic = deterministic

    def forward(
        self,
        Q: torch.Tensor,
        K: torch.Tensor,
        V: torch.Tensor,
        Bias1: torch.Tensor,
        Bias2: torch.Tensor,
    ) -> torch.Tensor:
        """
        Args:
            Q: (B, S, S, H, D)
            K: (B, S, S, H, D)
            V: (B, S, S, H, D)
            Bias1: (B, S, 1, 1, S)
            Bias2: (B, 1, H, S, S)
        Returns:
            output: (B, S, S, H, D)
        """
        output = TriAttentionFunction.apply(Q, K, V, Bias1, Bias2, self.deterministic)
        return output



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from typing import Any

import torch
import triton

from protenix.utils.distributed import DIST_WRAPPER

device_capability = torch.cuda.get_device_capability()
device_capability = f"{device_capability[0]}-{device_capability[1]}"

device_name = torch.cuda.get_device_name().replace(" ", "-")


def get_config_dir() -> Path:
    current_dir = Path(__file__).parent
    tri_cache_root = current_dir / "TriAttentionCache"
    rank = f"{DIST_WRAPPER.rank}"
    target_dir = tri_cache_root / rank
    if not target_dir.exists():
        target_dir.mkdir(parents=True, exist_ok=False)
        precache_dir = tri_cache_root / "0"
        if precache_dir.exists() and any(precache_dir.iterdir()):
            ret = os.system(f"cp {precache_dir}/* {target_dir}")
    return target_dir


config_dir = get_config_dir()


def config_to_dict(config: triton.Config) -> dict[str, Any]:
    # This assume we are not making use of `pre_hook` in the `triton.Config`
    return {
        "kwargs": config.kwargs,
        "num_warps": config.num_warps,
        "num_stages": config.num_stages,
    }


def dict_to_config(d: dict[str, Any]) -> triton.Config:
    return triton.Config(
        kwargs=d["kwargs"],
        num_warps=d["num_warps"],
        num_stages=d["num_stages"],
    )


# Base configs that should be ~ok for things <= 512.
_attention_fwd_configs = [
    triton.Config(kwargs={"BLOCK_M": 16, "BLOCK_N": 32}, num_warps=1, num_stages=2),
    triton.Config(kwargs={"BLOCK_M": 32, "BLOCK_N": 16}, num_warps=1, num_stages=2),
    triton.Config(kwargs={"BLOCK_M": 64, "BLOCK_N": 32}, num_warps=4, num_stages=2),
    triton.Config(kwargs={"BLOCK_M": 64, "BLOCK_N": 16}, num_warps=2, num_stages=3),
    triton.Config(kwargs={"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=1, num_stages=1),
    triton.Config(kwargs={"BLOCK_M": 128, "BLOCK_N": 16}, num_warps=2, num_stages=2),
    triton.Config(kwargs={"BLOCK_M": 128, "BLOCK_N": 32}, num_warps=2, num_stages=2),
    triton.Config(kwargs={"BLOCK_M": 32, "BLOCK_N": 64}, num_warps=2, num_stages=2),
    triton.Config(kwargs={"BLOCK_M": 64, "BLOCK_N": 32}, num_warps=2, num_stages=3),
    triton.Config(kwargs={"BLOCK_M": 32, "BLOCK_N": 16}, num_warps=1, num_stages=4),
]

_attention_bwd_dq_configs = [
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 32}, num_warps=2, num_stages=1),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 16}, num_warps=1, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 64}, num_warps=2, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=1, num_stages=1),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 16}, num_warps=2, num_stages=3),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 16}, num_warps=1, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 16}, num_warps=1, num_stages=3),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 64}, num_warps=1, num_stages=1),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 64}, num_warps=2, num_stages=1),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 32}, num_warps=2, num_stages=2),
]

_attention_bwd_dkdv_configs = [
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=1, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 16}, num_warps=2, num_stages=1),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 16}, num_warps=2, num_stages=1),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=2, num_stages=1),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=1, num_stages=3),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 32}, num_warps=2, num_stages=1),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 16}, num_warps=2, num_stages=1),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 16}, num_warps=1, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=1, num_stages=1),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 32}, num_warps=2, num_stages=1),
]

_attention_bwd_dbias2_configs = [
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 16}, num_warps=2, num_stages=2),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 16}, num_warps=2, num_stages=3),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 32}, num_warps=2, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=1, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 64}, num_warps=2, num_stages=2),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 64}, num_warps=2, num_stages=2),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 16}, num_warps=2, num_stages=1),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=2, num_stages=2),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 16}, num_warps=4, num_stages=3),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 32}, num_warps=2, num_stages=3),
]

_attention_bwd_preprocess_configs = [
    triton.Config({"BLOCK_M": 16}, num_warps=2, num_stages=2),
    triton.Config({"BLOCK_M": 16}, num_warps=2, num_stages=3),
    triton.Config({"BLOCK_M": 16}, num_warps=4, num_stages=4),
    triton.Config({"BLOCK_M": 16}, num_warps=4, num_stages=5),
    triton.Config({"BLOCK_M": 32}, num_warps=2, num_stages=2),
    triton.Config({"BLOCK_M": 32}, num_warps=2, num_stages=3),
    triton.Config({"BLOCK_M": 32}, num_warps=4, num_stages=4),
    triton.Config({"BLOCK_M": 32}, num_warps=4, num_stages=5),
]



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from torch_scatter: https://github.com/rusty1s/pytorch_scatter/tree/master

from typing import Optional

import torch


def broadcast(src: torch.Tensor, other: torch.Tensor, dim: int) -> torch.Tensor:
    if dim < 0:
        dim = other.dim() + dim
    if src.dim() == 1:
        for _ in range(0, dim):
            src = src.unsqueeze(0)
    for _ in range(src.dim(), other.dim()):
        src = src.unsqueeze(-1)
    src = src.expand(other.size())
    return src


def scatter_sum(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: Optional[torch.Tensor] = None,
    dim_size: Optional[int] = None,
) -> torch.Tensor:
    index = broadcast(index, src, dim)
    if out is None:
        size = list(src.size())
        if dim_size is not None:
            size[dim] = dim_size
        elif index.numel() == 0:
            size[dim] = 0
        else:
            size[dim] = int(index.max()) + 1
        out = torch.zeros(size, dtype=src.dtype, device=src.device)
        return out.scatter_add_(dim, index, src)
    else:
        return out.scatter_add_(dim, index, src)


def scatter_add(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: Optional[torch.Tensor] = None,
    dim_size: Optional[int] = None,
) -> torch.Tensor:
    return scatter_sum(src, index, dim, out, dim_size)


def scatter_mul(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: Optional[torch.Tensor] = None,
    dim_size: Optional[int] = None,
) -> torch.Tensor:
    return torch.ops.torch_scatter.scatter_mul(src, index, dim, out, dim_size)


def scatter_mean(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: Optional[torch.Tensor] = None,
    dim_size: Optional[int] = None,
) -> torch.Tensor:
    out = scatter_sum(src, index, dim, out, dim_size)
    dim_size = out.size(dim)

    index_dim = dim
    if index_dim < 0:
        index_dim = index_dim + src.dim()
    if index.dim() <= index_dim:
        index_dim = index.dim() - 1

    ones = torch.ones(index.size(), dtype=src.dtype, device=src.device)
    count = scatter_sum(ones, index, index_dim, None, dim_size)
    count[count < 1] = 1
    count = broadcast(count, out, dim)
    if out.is_floating_point():
        out.true_divide_(count)
    else:
        out.div_(count, rounding_mode="floor")
    return out


def scatter_min(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: Optional[torch.Tensor] = None,
    dim_size: Optional[int] = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    return torch.ops.torch_scatter.scatter_min(src, index, dim, out, dim_size)


def scatter_max(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: Optional[torch.Tensor] = None,
    dim_size: Optional[int] = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    return torch.ops.torch_scatter.scatter_max(src, index, dim, out, dim_size)


def scatter(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: Optional[torch.Tensor] = None,
    dim_size: Optional[int] = None,
    reduce: str = "sum",
) -> torch.Tensor:
    r"""
    |

    .. image:: https://raw.githubusercontent.com/rusty1s/pytorch_scatter/
            master/docs/source/_figures/add.svg?sanitize=true
        :align: center
        :width: 400px

    |

    Reduces all values from the :attr:`src` tensor into :attr:`out` at the
    indices specified in the :attr:`index` tensor along a given axis
    :attr:`dim`.
    For each value in :attr:`src`, its output index is specified by its index
    in :attr:`src` for dimensions outside of :attr:`dim` and by the
    corresponding value in :attr:`index` for dimension :attr:`dim`.
    The applied reduction is defined via the :attr:`reduce` argument.

    Formally, if :attr:`src` and :attr:`index` are :math:`n`-dimensional
    tensors with size :math:`(x_0, ..., x_{i-1}, x_i, x_{i+1}, ..., x_{n-1})`
    and :attr:`dim` = `i`, then :attr:`out` must be an :math:`n`-dimensional
    tensor with size :math:`(x_0, ..., x_{i-1}, y, x_{i+1}, ..., x_{n-1})`.
    Moreover, the values of :attr:`index` must be between :math:`0` and
    :math:`y - 1`, although no specific ordering of indices is required.
    The :attr:`index` tensor supports broadcasting in case its dimensions do
    not match with :attr:`src`.

    For one-dimensional tensors with :obj:`reduce="sum"`, the operation
    computes

    .. math::
        \mathrm{out}_i = \mathrm{out}_i + \sum_j~\mathrm{src}_j

    where :math:`\sum_j` is over :math:`j` such that
    :math:`\mathrm{index}_j = i`.

    .. note::

        This operation is implemented via atomic operations on the GPU and is
        therefore **non-deterministic** since the order of parallel operations
        to the same value is undetermined.
        For floating-point variables, this results in a source of variance in
        the result.

    :param src: The source tensor.
    :param index: The indices of elements to scatter.
    :param dim: The axis along which to index. (default: :obj:`-1`)
    :param out: The destination tensor.
    :param dim_size: If :attr:`out` is not given, automatically create output
        with size :attr:`dim_size` at dimension :attr:`dim`.
        If :attr:`dim_size` is not given, a minimal sized output tensor
        according to :obj:`index.max() + 1` is returned.
    :param reduce: The reduce operation (:obj:`"sum"`, :obj:`"mul"`,
        :obj:`"mean"`, :obj:`"min"` or :obj:`"max"`). (default: :obj:`"sum"`)

    :rtype: :class:`Tensor`

    .. code-block:: python

        from torch_scatter import scatter

        src = torch.randn(10, 6, 64)
        index = torch.tensor([0, 1, 0, 1, 2, 1])

        # Broadcasting in the first and last dim.
        out = scatter(src, index, dim=1, reduce="sum")

        print(out.size())

    .. code-block::

        torch.Size([10, 3, 64])
    """
    if reduce == "sum" or reduce == "add":
        return scatter_sum(src, index, dim, out, dim_size)
    if reduce == "mul":
        return scatter_mul(src, index, dim, out, dim_size)
    elif reduce == "mean":
        return scatter_mean(src, index, dim, out, dim_size)
    elif reduce == "min":
        return scatter_min(src, index, dim, out, dim_size)[0]
    elif reduce == "max":
        return scatter_max(src, index, dim, out, dim_size)[0]
    else:
        raise ValueError



from .infer_dataloader import get_inference_dataloader
from .json_maker import cif_to_input_json, merge_covalent_bonds
from .json_parser import (
    add_entity_atom_array,
    lig_file_to_atom_info,
    remove_leaving_atoms,
)
from .json_to_feature import SampleDictToFeatures

__all__ = [
    "get_inference_dataloader",
    "cif_to_input_json",
    "merge_covalent_bonds",
    "add_entity_atom_array",
    "lig_file_to_atom_info",
    "remove_leaving_atoms",
    "SampleDictToFeatures",
]



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0114,C0301
import os
from copy import deepcopy
from pathlib import Path

from protenix.config.extend_types import GlobalConfigValue, ListValue

PROTENIX_ROOT_DIR = os.environ.get("PROTENIX_ROOT_DIR", str(Path.home()))

default_test_configs = {
    "sampler_configs": {
        "sampler_type": "uniform",
    },
    "cropping_configs": {
        "method_weights": [
            0.0,  # ContiguousCropping
            0.0,  # SpatialCropping
            1.0,  # SpatialInterfaceCropping
        ],
        "crop_size": -1,
    },
    "lig_atom_rename": GlobalConfigValue("test_lig_atom_rename"),
    "shuffle_mols": GlobalConfigValue("test_shuffle_mols"),
    "shuffle_sym_ids": GlobalConfigValue("test_shuffle_sym_ids"),
    "constraint": {
        "enable": False,
        "fix_seed": False,  # True means use use the same contact in each evaluation.
    },
}

default_weighted_pdb_configs = {
    "sampler_configs": {
        "sampler_type": "weighted",
        "beta_dict": {
            "chain": 0.5,
            "interface": 1,
        },
        "alpha_dict": {
            "prot": 3,
            "nuc": 3,
            "ligand": 1,
        },
        "force_recompute_weight": True,
    },
    "cropping_configs": {
        "method_weights": ListValue([0.2, 0.4, 0.4]),
        "crop_size": GlobalConfigValue("train_crop_size"),
    },
    "sample_weight": 0.5,
    "limits": -1,
    "lig_atom_rename": GlobalConfigValue("train_lig_atom_rename"),
    "shuffle_mols": GlobalConfigValue("train_shuffle_mols"),
    "shuffle_sym_ids": GlobalConfigValue("train_shuffle_sym_ids"),
    # If enabled, the training settings for different constraint types,
    # providing the model a certain proportion of constraints
    # that meet specific conditions.
    "constraint": {
        "enable": False,
        "fix_seed": False,
        "pocket": {
            "prob": 0.0,
            "size": 1 / 3,
            "spec_binder_chain": False,
            "max_distance_range": {"PP": ListValue([6, 20]), "LP": ListValue([6, 20])},
            "group": "complex",
            "distance_type": "center_atom",
        },
        "contact": {
            "prob": 0.0,
            "size": 1 / 3,
            "max_distance_range": {
                "PP": ListValue([6, 30]),
                "PL": ListValue([4, 10]),
            },
            "group": "complex",
            "distance_type": "center_atom",
        },
        "substructure": {
            "prob": 0.0,
            "size": 0.8,
            "mol_type_pairs": {
                "PP": 15,
                "PL": 10,
                "LP": 10,
            },
            "feature_type": "one_hot",
            "ratios": {
                "full": [
                    0.0,
                    0.5,
                    1.0,
                ],  # ratio options of full chain substructure constraint
                "partial": 0.3,  # ratio of partial chain substructure constraint
            },
            "coord_noise_scale": 0.05,
            "spec_asym_id": False,
        },
        "contact_atom": {
            "prob": 0.0,
            "size": 1 / 3,
            "max_distance_range": {
                "PP": ListValue([2, 12]),
                "PL": ListValue([2, 8]),
            },
            "min_distance": -1,
            "group": "complex",
            "distance_type": "atom",
            "feature_type": "continuous",
        },
    },
}


data_configs = {
    "num_dl_workers": 16,
    "epoch_size": 10000,
    "train_ref_pos_augment": True,
    "test_ref_pos_augment": True,
    "train_sets": ListValue(["weightedPDB_before2109_wopb_nometalc_0925"]),
    "train_sampler": {
        "train_sample_weights": ListValue([1.0]),
        "sampler_type": "weighted",
    },
    "test_sets": ListValue(["recentPDB_1536_sample384_0925"]),
    "weightedPDB_before2109_wopb_nometalc_0925": {
        "base_info": {
            "mmcif_dir": os.path.join(PROTENIX_ROOT_DIR, "mmcif"),
            "bioassembly_dict_dir": os.path.join(
                PROTENIX_ROOT_DIR, "mmcif_bioassembly"
            ),
            "indices_fpath": os.path.join(
                PROTENIX_ROOT_DIR,
                "indices/weightedPDB_indices_before_2021-09-30_wo_posebusters_resolution_below_9.csv.gz",
            ),
            "pdb_list": "",
            "random_sample_if_failed": True,
            "max_n_token": -1,  # can be used for removing data with too many tokens.
            "use_reference_chains_only": False,
            "exclusion": {  # do not sample the data based on ions.
                "mol_1_type": ListValue(["ions"]),
                "mol_2_type": ListValue(["ions"]),
            },
        },
        **deepcopy(default_weighted_pdb_configs),
    },
    "recentPDB_1536_sample384_0925": {
        "base_info": {
            "mmcif_dir": os.path.join(PROTENIX_ROOT_DIR, "mmcif"),
            "bioassembly_dict_dir": os.path.join(
                PROTENIX_ROOT_DIR, "recentPDB_bioassembly"
            ),
            "indices_fpath": os.path.join(
                PROTENIX_ROOT_DIR, "indices/recentPDB_low_homology_maxtoken1536.csv"
            ),
            "pdb_list": os.path.join(
                PROTENIX_ROOT_DIR,
                "indices/recentPDB_low_homology_maxtoken1024_sample384_pdb_id.txt",
            ),
            "max_n_token": GlobalConfigValue("test_max_n_token"),  # filter data
            "sort_by_n_token": False,
            "group_by_pdb_id": True,
            "find_eval_chain_interface": True,
        },
        **deepcopy(default_test_configs),
    },
    "posebusters_0925": {
        "base_info": {
            "mmcif_dir": os.path.join(PROTENIX_ROOT_DIR, "posebusters_mmcif"),
            "bioassembly_dict_dir": os.path.join(
                PROTENIX_ROOT_DIR, "posebusters_bioassembly"
            ),
            "indices_fpath": os.path.join(
                PROTENIX_ROOT_DIR, "indices/posebusters_indices_mainchain_interface.csv"
            ),
            "pdb_list": "",
            "find_pocket": True,
            "find_all_pockets": False,
            "max_n_token": GlobalConfigValue("test_max_n_token"),  # filter data
        },
        **deepcopy(default_test_configs),
    },
    "msa": {
        "enable_prot_msa": True,
        "prot_seq_or_filename_to_msadir_jsons": ListValue(
            [os.path.join(PROTENIX_ROOT_DIR, "common/seq_to_pdb_index.json")]
        ),
        "prot_msadir_raw_paths": ListValue(
            [os.path.join(PROTENIX_ROOT_DIR, "mmcif_msa_template")]
        ),
        "prot_pairing_dbs": ListValue(["pairing"]),
        "prot_non_pairing_dbs": ListValue(
            ["pairing-non_pairing"]
        ),  # Separated by "-", "pairing-non_pairing" means both pairing and non_pairing are used as non_pairing,
        # with pairing used first.
        "prot_indexing_methods": ListValue(["sequence"]),
        "enable_rna_msa": True,  # enable rna msa
        "rna_seq_or_filename_to_msadir_jsons": ListValue(
            [os.path.join(PROTENIX_ROOT_DIR, "rna_msa/rna_sequence_to_pdb_chains.json")]
        ),
        "rna_msadir_raw_paths": ListValue(
            [os.path.join(PROTENIX_ROOT_DIR, "rna_msa/msas")]
        ),
        "rna_indexing_methods": ListValue(["sequence"]),
        "min_size": {
            "train": 1,
            "test": 1,
        },
        "max_size": {
            "train": 16384,
            "test": 16384,
        },
        "sample_cutoff": {
            "train": 16384,
            "test": 16384,
        },
    },
    "template": {
        "enable_prot_template": True,
        "template_dropout_rate": 0.0,
        "prot_template_mmcif_dir": os.path.join(PROTENIX_ROOT_DIR, "mmcif"),
        "prot_template_cache_dir": "",
        "prot_template_raw_paths": ListValue(
            [os.path.join(PROTENIX_ROOT_DIR, "mmcif_msa_template")]
        ),
        "prot_seq_or_filename_to_templatedir_jsons": ListValue(
            [os.path.join(PROTENIX_ROOT_DIR, "common/seq_to_pdb_index.json")]
        ),
        "prot_indexing_methods": ListValue(["sequence"]),
        "release_dates_path": os.path.join(
            PROTENIX_ROOT_DIR, "common/release_date_cache.json"
        ),
        "obsolete_pdbs_path": os.path.join(
            PROTENIX_ROOT_DIR, "common/obsolete_to_successor.json"
        ),
        "kalign_binary_path": "/usr/bin/kalign",  # apt-get install kalign
    },
    "ccd_components_file": os.path.join(PROTENIX_ROOT_DIR, "common/components.cif"),
    "ccd_components_rdkit_mol_file": os.path.join(
        PROTENIX_ROOT_DIR, "common/components.cif.rdkit_mol.pkl"
    ),
    "obsolete_release_data_csv": os.path.join(
        PROTENIX_ROOT_DIR, "common/obsolete_release_date.csv"
    ),
    "pdb_cluster_file": os.path.join(
        PROTENIX_ROOT_DIR, "common/clusters-by-entity-40.txt"
    ),
}



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn

from protenix.model.modules.primitives import Linear


# Adapted From openfold.model.heads
class DistogramHead(nn.Module):
    """Implements Algorithm 1 [Line17] in AF3

    Computes a distogram probability distribution.
    For use in computation of distogram loss, subsection 1.9.8 (AF2)

    Args:
        c_z (int, optional): hidden dim [for pair embedding]. Defaults to 128.
        no_bins (int, optional): Number of distogram bins. Defaults to 64.
    """

    def __init__(self, c_z: int = 128, no_bins: int = 64) -> None:
        super(DistogramHead, self).__init__()

        self.c_z = c_z
        self.no_bins = no_bins

        self.linear = Linear(
            in_features=self.c_z, out_features=self.no_bins, initializer="zeros"
        )

    def forward(self, z: torch.Tensor) -> torch.Tensor:  # [*, N, N, C_z]
        """
        Args:
            z (torch.Tensor): pair embedding
                [*, N_token, N_token, C_z]

        Returns:
            torch.Tensor: distogram probability distribution
                [*, N_token, N_token, no_bins]
        """
        # [*, N, N, no_bins]
        logits = self.linear(z)
        logits = logits + logits.transpose(-2, -3)
        return logits



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import torch


class EMAWrapper(object):
    """
    A wrapper class for exponential moving average of model weights.

    Args:
        model (torch.nn.Module): The PyTorch model to apply EMA to.
        decay (float): The decay rate for exponential moving average.
        mutable_param_keywords (Optional[List[str]]): List of keywords to identify parameters
            that should be updated by EMA. If None, all parameters are updated.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        decay: float = 0.999,
        mutable_param_keywords: Optional[List[str]] = None,
    ):
        self.model = model
        self.decay = decay
        if mutable_param_keywords is not None:
            self.mutable_param_keywords = [
                s.strip() for s in mutable_param_keywords if s.strip()
            ]
        else:
            self.mutable_param_keywords = None
        self.shadow = {}
        self.backup = {}

    def register(self) -> None:
        """
        Register the model's parameters and create initial shadow copies.
        """
        for name, param in self.model.named_parameters():
            self.shadow[name] = param.data.clone()

    def update(self) -> None:
        """
        Update the shadow copies of the parameters using the current model weights and decay rate.
        """
        for name, param in self.model.named_parameters():
            if self.mutable_param_keywords and not any(
                [keyword in name for keyword in self.mutable_param_keywords]
            ):
                continue
            assert name in self.shadow
            new_average = (1.0 - self.decay) * param.data + self.decay * self.shadow[
                name
            ]
            self.shadow[name] = new_average.clone()

    def apply_shadow(self) -> None:
        """
        Apply the shadow copies (EMA weights) to the model, backing up current weights.
        """
        for name, param in self.model.named_parameters():
            assert name in self.shadow
            self.backup[name] = param.data
            param.data = self.shadow[name]

    def restore(self) -> None:
        """
        Restore the original model weights from the backup.
        """
        for name, param in self.model.named_parameters():
            assert name in self.backup
            param.data = self.backup[name]
        self.backup = {}



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import time
from datetime import datetime, timedelta
from os.path import exists as opexists, join as opjoin
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
from biotite.structure import AtomArray
from typing_extensions import Self, TypeAlias

from protenix.data.constants import (
    DNA_CHAIN,
    LIGAND_CHAIN_TYPES,
    PROTEIN_CHAIN,
    RNA_CHAIN,
)
from protenix.data.msa.msa_utils import map_to_standard
from protenix.data.template.template_parser import HHRParser, HmmsearchA3MParser
from protenix.data.template.template_utils import (
    DAYS_BEFORE_QUERY_DATE,
    DistogramFeaturesConfig,
    TEMPLATE_FEATURES,
    TemplateFeatures,
    TemplateHitFeaturizer,
)
from protenix.data.utils import pad_to
from protenix.utils.file_io import load_json_cached
from protenix.utils.logger import get_logger

logger = get_logger(__name__)

BatchDict: TypeAlias = dict[str, np.ndarray]
FeatureDict: TypeAlias = Mapping[str, np.ndarray]


class TemplateSourceManager:
    """
    Manages template data retrieval and loading from multiple sources.

    Args:
        raw_paths: List of base paths for template storage.
        indexing_methods: List of indexing methods (e.g., 'sequence' or 'pdb_id').
        mappings: Dictionary mapping source index to its respective lookup table.
        enabled: Whether template loading is enabled.
    """

    def __init__(
        self,
        raw_paths: Sequence[str],
        indexing_methods: Sequence[str],
        mappings: Dict[int, Dict[str, Any]],
        enabled: bool = True,
    ) -> None:
        self.raw_paths = raw_paths
        self.indexing_methods = indexing_methods
        self.mappings = mappings
        self.enabled = enabled

    def fetch_template_paths(
        self, pdb_id: str, query_sequence: str, chain_entity_type: str
    ) -> List[str]:
        """
        Fetches template file paths from the configured sources.

        Args:
            pdb_id: PDB identifier of the query.
            query_sequence: Query sequence.
            chain_entity_type: Type of the chain (e.g., PROTEIN_CHAIN).

        Returns:
            A list of paths to found template files (.a3m or .hhr).
        """
        if not self.enabled or chain_entity_type != PROTEIN_CHAIN:
            return []

        template_paths = []
        for i, (path, method) in enumerate(zip(self.raw_paths, self.indexing_methods)):
            mapping = self.mappings.get(i, {})
            key = pdb_id if method == "pdb_id" else query_sequence

            if key not in mapping:
                continue

            dir_path = opjoin(path, str(mapping[key]))
            # Check for multiple possible template filenames
            possible_subpaths = [
                "hmmsearch.a3m",
                "concat.hhr",
            ]
            for subpath in possible_subpaths:
                full_path = opjoin(dir_path, subpath)
                if opexists(full_path):
                    template_paths.append(full_path)
        return template_paths


class TemplateFeatureAssemblyLine:
    """
    Orchestrates the conversion of raw templates into finalized Protenix features.

    Args:
        max_templates: Maximum number of templates to include in the features.
    """

    def __init__(self, max_templates: int = 4) -> None:
        self.max_templates = max_templates

    def assemble(
        self,
        bioassembly: Mapping[int, Mapping[str, Any]],
        standard_token_idxs: np.ndarray,
    ) -> "Templates":
        """
        Executes the complete feature assembly pipeline.

        Args:
            bioassembly: Mapping of asymmetric IDs to chain information.
            standard_token_idxs: Array of standardized residue indices.

        Returns:
            An assembled Templates object.
        """
        np_chains_list = []
        polymer_entity_features = {True: {}, False: {}}
        # Identify entities where template features can be safely copied (same sequence)
        safe_entity_ids = get_safe_entity_id_for_template_copy(bioassembly)

        for asym_id, info in bioassembly.items():
            chain_id = info["chain_id"]
            entity_id = info["entity_id"]
            chain_type = info["chain_entity_type"]
            num_tokens = len(info["sequence"])

            # Templates are currently only supported for protein chains with sufficient length
            skip_chain = chain_type != PROTEIN_CHAIN or num_tokens <= 4

            if (entity_id not in polymer_entity_features[skip_chain]) or (
                entity_id not in safe_entity_ids
            ):
                templates = info["templates"]
                if skip_chain or not templates:
                    template_features = TemplateFeatures.empty_template_features(
                        num_tokens
                    )
                else:
                    # Package and fix template features
                    template_features = TemplateFeatures.package_template_features(
                        hit_features=templates
                    )
                    template_features = TemplateFeatures.fix_template_features(
                        template_features=template_features,
                        num_res=num_tokens,
                    )
                # Reduce to requested maximum number of templates
                template_features = _reduce_template_features(
                    template_features, self.max_templates
                )
                if entity_id in safe_entity_ids:
                    polymer_entity_features[skip_chain][entity_id] = template_features

            if entity_id in safe_entity_ids:
                feats = polymer_entity_features[skip_chain][entity_id].copy()
            else:
                feats = template_features

            feats["chain_id"] = chain_id
            np_chains_list.append(feats)

        # Pad the number of templates to max_templates for each chain to allow concatenation
        for chain in np_chains_list:
            chain["template_aatype"] = pad_to(
                chain["template_aatype"], (self.max_templates, None)
            )
            chain["template_atom_positions"] = pad_to(
                chain["template_atom_positions"],
                (self.max_templates, None, None, None),
            )
            chain["template_atom_mask"] = pad_to(
                chain["template_atom_mask"], (self.max_templates, None, None)
            )

        # Concatenate features along the residue dimension
        merged_example = {
            ft: np.concatenate([c[ft] for c in np_chains_list], axis=1)
            for ft in np_chains_list[0]
            if ft in TEMPLATE_FEATURES
        }

        # Crop/index merged features using standard token indices
        for feature_name, v in merged_example.items():
            merged_example[feature_name] = v[
                : self.max_templates, standard_token_idxs, ...
            ]

        return Templates(
            aatype=merged_example["template_aatype"],
            atom_positions=merged_example["template_atom_positions"],
            atom_mask=merged_example["template_atom_mask"].astype(bool),
        )


class TemplateFeaturizer:
    """
    Main class for template featurization in training and evaluation.

    Args:
        dataset_name: Name of the dataset.
        prot_template_mmcif_dir: Directory containing template mmCIF files.
        prot_template_cache_dir: Directory for cached template data.
        prot_template_raw_paths: Base paths for protein template storage.
        prot_seq_or_filename_to_templatedir_jsons: JSON files mapping sequences/PDB IDs to template directories.
        prot_indexing_methods: Indexing methods for each source.
        enable_prot_template: Whether to enable protein templates.
        template_dropout_rate: Dropout rate for templates during training.
        stage: 'train' or 'eval'.
        kalign_binary_path: Path to kalign binary.
        release_dates_path: Path to PDB release dates JSON.
        obsolete_pdbs_path: Path to obsolete PDBs mapping JSON.
        **kwargs: Additional arguments.
    """

    def __init__(
        self,
        dataset_name: str = "",
        prot_template_mmcif_dir: str = "",
        prot_template_cache_dir: str = None,
        prot_template_raw_paths: Sequence[str] = [""],
        prot_seq_or_filename_to_templatedir_jsons: Sequence[str] = [""],
        prot_indexing_methods: Sequence[str] = ["sequence"],
        enable_prot_template: bool = True,
        template_dropout_rate: float = 0.0,
        stage: str = "train",
        kalign_binary_path: str = "",
        release_dates_path: str = "",
        obsolete_pdbs_path: str = "",
        **kwargs,
    ) -> None:
        # Validate inputs
        assert len(prot_template_raw_paths) == len(
            prot_seq_or_filename_to_templatedir_jsons
        ), "Mismatch between raw paths and JSON mapping files."
        assert len(kalign_binary_path) > 0, "kalign_binary_path must be provided."
        assert len(release_dates_path) > 0, "release_dates_path must be provided."
        assert len(obsolete_pdbs_path) > 0, "obsolete_pdbs_path must be provided."
        assert len(prot_template_raw_paths) == len(
            prot_indexing_methods
        ), "Mismatch between raw paths and indexing methods."

        # Currently, multiple source training strategy is not determined
        assert (
            len(prot_template_raw_paths) == 1
        ), "Only single source template supported."

        if stage != "train":
            max_template_date = "2021-09-30"
        else:
            if ("distillation" in dataset_name) or ("openprotein" in dataset_name):
                max_template_date = "2018-04-30"
            else:
                max_template_date = "3000-01-01"

        self.stage = stage
        self.max_template_date = datetime.strptime(max_template_date, "%Y-%m-%d")
        self.dataset_name = dataset_name
        self.enable_prot_template = enable_prot_template
        self.template_dropout_rate = template_dropout_rate

        # Initialize source manager
        mappings = {
            i: load_json_cached(p)
            for i, p in enumerate(prot_seq_or_filename_to_templatedir_jsons)
        }
        self.source_mgr = TemplateSourceManager(
            raw_paths=prot_template_raw_paths,
            indexing_methods=prot_indexing_methods,
            mappings=mappings,
            enabled=enable_prot_template,
        )

        # Initialize online featurizer
        self.online_featurizer = TemplateHitFeaturizer(
            mmcif_dir=prot_template_mmcif_dir,
            template_cache_dir=prot_template_cache_dir,
            max_hits=None if self.stage == "train" else 4,
            kalign_binary_path=kalign_binary_path,
            max_template_date=max_template_date,
            release_dates_path=release_dates_path,
            obsolete_pdbs_path=obsolete_pdbs_path,
            _shuffle_top_k_prefiltered=20 if self.stage == "train" else None,
        )

        self._last_profile: Dict[str, Any] = {
            "elapsed_seconds": 0.0,
            "raw_templates": 0,
            "selected_templates": 0,
            "chain_profiles": [],
        }

    def get_template(
        self,
        pdb_id: str,
        query_sequence: str,
        sequence_uid: str,
        query_release_date: Optional[datetime],
        chain_entity_type: str,
    ) -> Tuple[List[Dict[str, Any]], int, Dict[str, float]]:
        """
        Fetches and processes templates for a single sequence.

        Args:
            pdb_id: PDB identifier.
            query_sequence: Query sequence string.
            sequence_uid: Unique identifier for the sequence.
            query_release_date: Release date of the query structure.
            chain_entity_type: Type of the chain.

        Returns:
            A tuple of (template_features_list, raw_hit_count, timing_track).
        """
        if not self.enable_prot_template or chain_entity_type != PROTEIN_CHAIN:
            return [], 0, {}

        # Fetch template paths from sources
        template_paths = self.source_mgr.fetch_template_paths(
            pdb_id, query_sequence, chain_entity_type
        )

        hmmsearched_a3m, hmmsearched_hhr = "", ""
        for path in template_paths:
            with open(path, "r") as f:
                content = f.read()
            if not content:
                continue
            if path.endswith(".a3m"):
                hmmsearched_a3m += content
            elif path.endswith(".hhr"):
                hmmsearched_hhr += content

        # Determine cutoff date for filtering templates
        if query_release_date is None:
            cutoff_date = self.max_template_date
        else:
            cutoff_date = min(
                self.max_template_date,
                query_release_date - timedelta(days=DAYS_BEFORE_QUERY_DATE),
            )

        # Parse hits from a3m or hhr content
        if hmmsearched_hhr:
            hits = HHRParser.parse(hhr_string=hmmsearched_hhr)
        else:
            hits = HmmsearchA3MParser.parse(
                query_seq=query_sequence, a3m_str=hmmsearched_a3m, skip_first=False
            )

        raw_hit_count = len(hits)
        if raw_hit_count == 0:
            return [], 0, {}

        # Process hits into features using the online featurizer
        result, time_track = self.online_featurizer.get_templates(
            sequence_uid=sequence_uid,
            query_sequence=query_sequence,
            hits=hits,
            max_template_date=cutoff_date,
        )
        return result.features, raw_hit_count, time_track

    def make_template_features(
        self,
        bioassembly_dict: Dict[str, Any],
        selected_token_indices: Optional[np.ndarray],
        entity_to_asym_id_int: Mapping[str, Sequence[int]],
    ) -> Dict[str, np.ndarray]:
        """
        Main method to generate template features for a bioassembly.

        Args:
            bioassembly_dict: Dictionary containing bioassembly data.
            selected_token_indices: Indices of selected tokens (for cropping).
            entity_to_asym_id_int: Mapping from entity ID to asymmetric IDs.

        Returns:
            Dictionary of processed template features.
        """
        profile_start = time.time()
        atom_array = bioassembly_dict["atom_array"]
        token_array = bioassembly_dict["token_array"]

        # Identify selected asymmetric IDs
        centre_atom_indices = (
            token_array[selected_token_indices]
            if selected_token_indices is not None
            else token_array
        ).get_annotation("centre_atom_index")
        selected_asym_ids = set(atom_array[centre_atom_indices].asym_id_int)

        # Build mapping from asym_id to entity_id and sequence
        asym_to_entity_id: Dict[int, str] = {}
        for eid, asyms in entity_to_asym_id_int.items():
            for aid in asyms:
                if aid in selected_asym_ids:
                    asym_to_entity_id[aid] = eid

        entity_id_to_sequence = {
            eid: bioassembly_dict["sequences"].get(eid)
            for eid in asym_to_entity_id.values()
        }

        # Patch sequences for ligands if missing
        asym_id_to_sequence = {}
        for aid, eid in asym_to_entity_id.items():
            seq = entity_id_to_sequence.get(eid)
            if seq is None:
                ligand_mask = atom_array.asym_id_int == aid
                assert atom_array.is_ligand[ligand_mask].all(), "Unknown molecule type"
                seq = "X" * ligand_mask.sum()
            asym_id_to_sequence[aid] = seq

        poly_types_mapping = {
            "polypeptide(L)": PROTEIN_CHAIN,
            "polyribonucleotide": RNA_CHAIN,
            "polydeoxyribonucleotide": DNA_CHAIN,
        }
        entity_poly_type = bioassembly_dict.get("entity_poly_type", {})

        query_release_date = None
        if self.stage == "train" and "release_date" in bioassembly_dict:
            query_release_date = datetime.strptime(
                bioassembly_dict["release_date"], "%Y-%m-%d"
            )

        # Global template dropout
        drop_template = False
        if self.stage == "train" and self.template_dropout_rate > 0.0:
            drop_template = np.random.uniform(0, 1) < self.template_dropout_rate

        template_meta_infos = {}
        seq_to_templates_cache: Dict[str, Tuple[List[Dict], int, Dict]] = {}
        profile = {
            "elapsed_seconds": 0.0,
            "raw_templates": 0,
            "selected_templates": 0,
            "chain_profiles": [],
        }

        pdb_id = bioassembly_dict.get("pdb_id", "unknown")

        for aid, sequence in asym_id_to_sequence.items():
            chain_start = time.time()
            eid = asym_to_entity_id[aid]
            chain_type = poly_types_mapping.get(
                entity_poly_type.get(eid, "non-polymer"), LIGAND_CHAIN_TYPES
            )

            if drop_template or chain_type != PROTEIN_CHAIN:
                templates, raw_hit_count, time_track = [], 0, {}
            else:
                if sequence in seq_to_templates_cache:
                    templates, raw_hit_count, time_track = seq_to_templates_cache[
                        sequence
                    ]
                else:
                    templates, raw_hit_count, time_track = self.get_template(
                        pdb_id=pdb_id,
                        query_sequence=sequence,
                        sequence_uid=f"{pdb_id}_{aid}",
                        query_release_date=query_release_date,
                        chain_entity_type=chain_type,
                    )
                    seq_to_templates_cache[sequence] = (
                        templates,
                        raw_hit_count,
                        time_track,
                    )

            chain_id = atom_array.chain_id[atom_array.asym_id_int == aid][0]
            if chain_type == PROTEIN_CHAIN:
                num_sel = len(templates) if isinstance(templates, list) else 0
                profile["raw_templates"] += raw_hit_count
                profile["selected_templates"] += num_sel
                profile["chain_profiles"].append(
                    {
                        "asym_id": aid,
                        "chain_id": chain_id,
                        "elapsed_seconds": time.time() - chain_start,
                        "raw_templates": raw_hit_count,
                        "selected_templates": num_sel,
                        "chain_entity_type": chain_type,
                        "time_track": time_track,
                    }
                )

            template_meta_infos[aid] = {
                "entity_id": eid,
                "chain_id": chain_id,
                "sequence": sequence,
                "chain_entity_type": chain_type,
                "templates": templates,
            }

        if not get_safe_entity_id_for_template_copy(template_meta_infos):
            logger.warning(
                f"PDB {pdb_id} has inconsistent sequences within same entity."
            )

        # Coordinate mapping to standardized indices
        ca = atom_array[centre_atom_indices]
        std_idxs = map_to_standard(ca.asym_id_int, ca.res_id, template_meta_infos)

        # Assemble final features
        max_t = 4 if self.stage != "train" else 4  # Default to 4
        assembly_line = TemplateFeatureAssemblyLine(max_templates=max_t)
        template_features = assembly_line.assemble(
            template_meta_infos, std_idxs
        ).as_protenix_dict()

        profile["elapsed_seconds"] = time.time() - profile_start
        self._last_profile = profile
        return template_features

    def __call__(
        self,
        bioassembly_dict: Dict[str, Any],
        selected_indices: np.ndarray,
        entity_to_asym_id_int: Mapping[str, Sequence[int]],
    ) -> Dict[str, np.ndarray]:
        """Call method proxying to make_template_features."""
        return self.make_template_features(
            bioassembly_dict, selected_indices, entity_to_asym_id_int
        )

    def get_last_profile(self) -> Dict[str, Union[float, int]]:
        """Returns profiling information for the last call."""
        return self._last_profile


@dataclasses.dataclass(frozen=True)
class Templates:
    """Dataclass containing template features."""

    # aatype: [num_templates, num_res]
    aatype: np.ndarray
    # atom_positions: [num_templates, num_res, 24, 3]
    atom_positions: np.ndarray
    # atom_mask: [num_templates, num_res, 24]
    atom_mask: np.ndarray

    @classmethod
    def from_data_dict(cls, batch: BatchDict) -> Self:
        """Construct instance from a data dictionary."""
        return cls(
            aatype=batch["template_aatype"],
            atom_positions=batch["template_atom_positions"],
            atom_mask=batch["template_atom_mask"],
        )

    def as_data_dict(self) -> BatchDict:
        """Convert to a standard data dictionary."""
        return {
            "template_aatype": self.aatype,
            "template_atom_positions": self.atom_positions,
            "template_atom_mask": self.atom_mask,
        }

    def as_protenix_dict(self) -> BatchDict:
        """Compute additional features and return as Protenix dictionary."""
        features = self.as_data_dict()
        dgrams, pb_masks = [], []
        unit_vectors, bb_masks = [], []

        num_templates = self.aatype.shape[0]
        for i in range(num_templates):
            aatype = self.aatype[i]
            mask = self.atom_mask[i]
            pos = self.atom_positions[i] * mask[..., None]

            # Compute pseudo-beta positions and mask
            pb_pos, pb_mask = TemplateFeatures.pseudo_beta_fn(aatype, pos, mask)
            pb_mask_2d = pb_mask[:, None] * pb_mask[None, :]

            # Compute distogram
            dgram = TemplateFeatures.dgram_from_positions(
                pb_pos,
                config=DistogramFeaturesConfig(
                    min_bin=3.25, max_bin=50.75, num_bins=39
                ),
            )
            dgrams.append(dgram * pb_mask_2d[..., None])
            pb_masks.append(pb_mask_2d)

            # Compute normalized unit vectors between residues
            uv, bb_mask_2d = TemplateFeatures.compute_template_unit_vector(
                aatype, pos, mask
            )
            unit_vectors.append(uv * bb_mask_2d[..., None])
            bb_masks.append(bb_mask_2d)

        features.update(
            {
                "template_pseudo_beta_mask": np.stack(pb_masks),
                "template_distogram": np.stack(dgrams),
                "template_unit_vector": np.stack(unit_vectors),
                "template_backbone_frame_mask": np.stack(bb_masks),
            }
        )
        return features


def _reduce_template_features(
    template_features: FeatureDict, max_templates: int
) -> FeatureDict:
    """Reduces templates to the requested maximum number."""
    num_t = template_features["template_aatype"].shape[0]
    keep_mask = np.arange(num_t) < max_templates
    fields = TEMPLATE_FEATURES + ("template_release_timestamp",)
    return {k: v[keep_mask] for k, v in template_features.items() if k in fields}


def get_safe_entity_id_for_template_copy(
    bioassembly: Mapping[int, Mapping[str, Any]],
) -> List[str]:
    """Identifies entity IDs that have consistent sequences across all chains."""
    eid_to_seqs = {}
    for aid, info in bioassembly.items():
        eid = info["entity_id"]
        seq = info["sequence"]
        eid_to_seqs.setdefault(eid, set()).add(seq)
    return [eid for eid, seqs in eid_to_seqs.items() if len(seqs) == 1]


class InferenceTemplateFeaturizer:
    """Simplified featurizer for inference, leveraging the same assembly logic."""

    @staticmethod
    def make_template_feature(
        bioassembly: Sequence[Mapping[str, Any]],
        atom_array: AtomArray,
        use_template: bool = True,
        online_template_featurizer: Optional[TemplateHitFeaturizer] = None,
    ) -> Dict[str, np.ndarray]:
        """
        Generates template features during inference.

        Args:
            bioassembly: List of entity information from the input JSON.
            atom_array: Parsed atom structure.
            use_template: Whether to use templates.
            online_template_featurizer: Featurizer for processing template hits.

        Returns:
            Dictionary of template features.
        """
        logger.info("Calling InferenceTemplateFeaturizer.make_template_feature")
        template_meta_infos = {}
        curr_asym_id = 0

        for eid, info in enumerate(bioassembly):
            seq, count, ctype, t_path = "", 0, LIGAND_CHAIN_TYPES, ""

            if "proteinChain" in info:
                c = info["proteinChain"]
                seq, count, ctype, t_path = (
                    c["sequence"],
                    c["count"],
                    PROTEIN_CHAIN,
                    c.get("templatesPath", ""),
                )
            elif "rnaSequence" in info:
                c = info["rnaSequence"]
                seq, count, ctype = c["sequence"], c["count"], RNA_CHAIN
            elif "dnaSequence" in info:
                c = info["dnaSequence"]
                seq, count, ctype = c["sequence"], c["count"], DNA_CHAIN
            elif "ligand" in info:
                count, ctype = info["ligand"]["count"], LIGAND_CHAIN_TYPES
                seq = "X" * (atom_array.asym_id_int == curr_asym_id).sum()

            templates = []
            if t_path and use_template and online_template_featurizer:
                assert ctype == PROTEIN_CHAIN, "Only protein templates are supported."
                with open(t_path, "r") as f:
                    content = f.read()

                if t_path.endswith(".hhr"):
                    hits = HHRParser.parse(hhr_string=content)
                elif t_path.endswith(".a3m"):
                    hits = HmmsearchA3MParser.parse(
                        query_seq=seq, a3m_str=content, skip_first=False
                    )
                else:
                    raise ValueError(f"Unsupported template format: {t_path}")

                result, _ = online_template_featurizer.get_templates(
                    sequence_uid=seq,
                    query_sequence=seq,
                    hits=hits,
                    max_template_date=None,
                )
                templates = result.features
                logger.info(f"Found {len(templates)} templates for sequence {seq}")

            for i in range(count):
                aid = curr_asym_id + i
                template_meta_infos[aid] = {
                    "entity_id": eid,
                    "chain_id": atom_array.chain_id[atom_array.asym_id_int == aid][0],
                    "sequence": seq,
                    "chain_entity_type": ctype,
                    "templates": templates,
                }
            curr_asym_id += count

        # Coordinate mapping
        ca = atom_array[atom_array.centre_atom_mask.astype(bool)]
        std_idxs = map_to_standard(ca.asym_id_int, ca.res_id, template_meta_infos)

        # Assemble features
        return (
            TemplateFeatureAssemblyLine(max_templates=4)
            .assemble(template_meta_infos, std_idxs)
            .as_protenix_dict()
        )



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from collections import defaultdict
from typing import Optional, Union

import numpy as np
import torch
from biotite.structure import Atom, AtomArray, get_residue_starts
from sklearn.neighbors import KDTree

from protenix.data.constants import get_all_elems, STD_RESIDUES, STD_RESIDUES_WITH_GAP
from protenix.data.tokenizer import Token, TokenArray
from protenix.data.utils import get_atom_level_token_mask, get_ligand_polymer_bond_mask
from protenix.utils.geometry import angle_3p, random_transform


class Featurizer(object):
    """
    Args:
        cropped_token_array (TokenArray): TokenArray object after cropping
        cropped_atom_array (AtomArray): AtomArray object after cropping
        ref_pos_augment (bool): Boolean indicating whether apply random rotation and translation on ref_pos
        lig_atom_rename (bool): Boolean indicating whether rename atom name for ligand atoms
        include_discont_poly_poly_bonds (bool): Boolean indicating whether
                                        include discontinuous polymer-polymer bonds
    """

    def __init__(
        self,
        cropped_token_array: TokenArray,
        cropped_atom_array: AtomArray,
        ref_pos_augment: bool = True,
        lig_atom_rename: bool = False,
        include_discont_poly_poly_bonds: bool = False,
    ) -> None:
        self.cropped_token_array = cropped_token_array

        self.cropped_atom_array = cropped_atom_array
        self.ref_pos_augment = ref_pos_augment
        self.lig_atom_rename = lig_atom_rename
        self.include_discont_poly_poly_bonds = include_discont_poly_poly_bonds

    @staticmethod
    def encoder(
        encode_def_dict_or_list: Optional[Union[dict, list[str]]], input_list: list[str]
    ) -> torch.Tensor:
        """
        Encode a list of input values into a binary format using a specified encoding definition list.

        Args:
            encode_def_dict_or_list (list or dict): A list or dict of encoding definitions.
            input_list (list): A list of input values to be encoded.

        Returns:
            torch.Tensor: A tensor representing the binary encoding of the input values.
        """
        num_keys = len(encode_def_dict_or_list)
        if isinstance(encode_def_dict_or_list, dict):
            items = encode_def_dict_or_list.items()
            assert (
                num_keys == max(encode_def_dict_or_list.values()) + 1
            ), "Do not use discontinuous number, which might causing potential bugs in the code"
        elif isinstance(encode_def_dict_or_list, list):
            items = ((key, idx) for idx, key in enumerate(encode_def_dict_or_list))
        else:
            raise TypeError(
                "encode_def_dict_or_list must be a list or dict, "
                f"but got {type(encode_def_dict_or_list)}"
            )
        onehot_dict = {
            key: [int(i == idx) for i in range(num_keys)] for key, idx in items
        }
        onehot_encoded_data = [onehot_dict[item] for item in input_list]
        onehot_tensor = torch.Tensor(onehot_encoded_data)
        return onehot_tensor

    @staticmethod
    def restype_onehot_encoded(restype_list: list[str]) -> torch.Tensor:
        """
        Ref: AlphaFold3 SI Table 5 "restype"
        One-hot encoding of the sequence. 32 possible values: 20 amino acids + unknown,
        4 RNA nucleotides + unknown, 4 DNA nucleotides + unknown, and gap.
        Ligands represented as “unknown amino acid”.

        Args:
            restype_list (List[str]): A list of residue types.
                                      The residue type of ligand should be "UNK" in the input list.

        Returns:
            torch.Tensor:  A Tensor of one-hot encoded residue types
        """

        return Featurizer.encoder(STD_RESIDUES_WITH_GAP, restype_list)

    @staticmethod
    def elem_onehot_encoded(elem_list: list[str]) -> torch.Tensor:
        """
        Ref: AlphaFold3 SI Table 5 "ref_element"
        One-hot encoding of the element atomic number for each atom
        in the reference conformer, up to atomic number 128.

        Args:
            elem_list (List[str]): A list of element symbols.

        Returns:
            torch.Tensor:  A Tensor of one-hot encoded elements
        """
        return Featurizer.encoder(get_all_elems(), elem_list)

    @staticmethod
    def ref_atom_name_chars_encoded(atom_names: list[str]) -> torch.Tensor:
        """
        Ref: AlphaFold3 SI Table 5 "ref_atom_name_chars"
        One-hot encoding of the unique atom names in the reference conformer.
        Each character is encoded as ord(c) − 32, and names are padded to length 4.

        Args:
            atom_name_list (List[str]): A list of atom names.

        Returns:
            torch.Tensor:  A Tensor of character encoded atom names
        """
        onehot_dict = {}
        for index, key in enumerate(range(64)):
            onehot = [0] * 64
            onehot[index] = 1
            onehot_dict[key] = onehot
        # [N_atom, 4, 64]
        mol_encode = []
        for atom_name in atom_names:
            # [4, 64]
            atom_encode = []
            for name_str in atom_name.ljust(4):
                atom_encode.append(onehot_dict[ord(name_str) - 32])
            mol_encode.append(atom_encode)
        onehot_tensor = torch.Tensor(mol_encode)
        return onehot_tensor

    @staticmethod
    def get_prot_nuc_frame(token: Token, centre_atom: Atom) -> tuple[int, list[int]]:
        """
        Ref: AlphaFold3 SI Chapter 4.3.2
        For proteins/DNA/RNA, we use the three atoms [N, CA, C] / [C1', C3', C4']

        Args:
            token (Token): Token object.
            centre_atom (Atom): Biotite Atom object of Token centre atom.

        Returns:
            has_frame (int): 1 if the token has frame, 0 otherwise.
            frame_atom_index (List[int]): The index of the atoms used to construct the frame.
        """
        if centre_atom.mol_type == "protein":
            # For protein
            abc_atom_name = ["N", "CA", "C"]
        else:
            # For DNA and RNA
            abc_atom_name = [r"C1'", r"C3'", r"C4'"]

        idx_in_atom_indices = []
        for i in abc_atom_name:
            if centre_atom.mol_type == "protein" and "N" not in token.atom_names:
                return 0, [-1, -1, -1]
            elif centre_atom.mol_type != "protein" and "C1'" not in token.atom_names:
                return 0, [-1, -1, -1]
            idx_in_atom_indices.append(token.atom_names.index(i))
        # Protein/DNA/RNA always has frame
        has_frame = 1
        frame_atom_index = [token.atom_indices[i] for i in idx_in_atom_indices]
        return has_frame, frame_atom_index

    @staticmethod
    def get_lig_frame(
        token: Token,
        centre_atom: Atom,
        lig_res_ref_conf_kdtree: dict[str, tuple[KDTree, list[int]]],
        ref_pos: torch.Tensor,
        ref_mask: torch.Tensor,
    ) -> tuple[int, list[int]]:
        """
        Ref: AlphaFold3 SI Chapter 4.3.2
        For ligands, we use the reference conformer of the ligand to construct the frame.

        Args:
            token (Token): Token object.
            centre_atom (Atom): Biotite Atom object of Token centre atom.
            lig_res_ref_conf_kdtree (Dict[str, Tuple[KDTree, List[int]]]): A dictionary of KDTree objects and atom indices.
            ref_pos (torch.Tensor): Atom positions in the reference conformer. Size=[N_atom, 3]
            ref_mask (torch.Tensor): Mask indicating which atom slots are used in the reference conformer. Size=[N_atom]

        Returns:
            tuple[int, List[int]]:
                has_frame (int): 1 if the token has frame, 0 otherwise.
                frame_atom_index (List[int]): The index of the atoms used to construct the frame.
        """
        kdtree, atom_ids = lig_res_ref_conf_kdtree[centre_atom.ref_space_uid]
        b_ref_pos = ref_pos[token.centre_atom_index]
        b_idx = token.centre_atom_index
        if kdtree is None:
            # Atom num < 3
            frame_atom_index = [-1, b_idx, -1]
            has_frame = 0
        else:
            _dist, ind = kdtree.query([b_ref_pos], k=3)
            a_idx, c_idx = atom_ids[ind[0][1]], atom_ids[ind[0][2]]
            frame_atom_index = [a_idx, b_idx, c_idx]

            # Check if reference confomrer vaild
            has_frame = all([ref_mask[idx] for idx in frame_atom_index])

            # Colinear check
            if has_frame:
                vec1 = ref_pos[frame_atom_index[1]] - ref_pos[frame_atom_index[0]]
                vec2 = ref_pos[frame_atom_index[2]] - ref_pos[frame_atom_index[1]]
                # ref_pos can be all zeros, in which case has_frame=0
                is_zero_norm = np.isclose(
                    np.linalg.norm(vec1, axis=-1), 0
                ) or np.isclose(np.linalg.norm(vec2, axis=-1), 0)
                if is_zero_norm:
                    has_frame = 0
                else:
                    theta_degrees = angle_3p(
                        *[ref_pos[idx] for idx in frame_atom_index]
                    )
                    is_colinear = theta_degrees <= 25 or theta_degrees >= 155
                    if is_colinear:
                        has_frame = 0
        return has_frame, frame_atom_index

    @staticmethod
    def get_token_frame(
        token_array: TokenArray,
        atom_array: AtomArray,
        ref_pos: torch.Tensor,
        ref_mask: torch.Tensor,
    ) -> TokenArray:
        """
        Ref: AlphaFold3 SI Chapter 4.3.2
        The atoms (a_i, b_i, c_i) used to construct token i’s frame depend on the chain type of i:
        Protein tokens use their residue’s backbone (N, Cα, C),
        while DNA and RNA tokens use (C1′, C3′, C4′) atoms of their residue.
        All other tokens (small molecules, glycans, ions) contain only one atom per token.
        The token atom is assigned to b_i, the closest atom to the token atom is a_i,
        and the second closest atom to the token atom is c_i.
        If this set of three atoms is close to colinear (less than 25 degree deviation),
        or if three atoms do not exist in the chain (e.g. a sodium ion),
        then the frame is marked as invalid.

        Note: frames constucted from reference conformer

        Args:
            token_array (TokenArray): A list of tokens.
            atom_array (AtomArray): An atom array.
            ref_pos (torch.Tensor): Atom positions in the reference conformer. Size=[N_atom, 3]
            ref_mask (torch.Tensor): Mask indicating which atom slots are used in the reference conformer. Size=[N_atom]

        Returns:
            TokenArray: A TokenArray with updated frame annotations.
                        - has_frame: 1 if the token has frame, 0 otherwise.
                        - frame_atom_index: The index of the atoms used to construct the frame.
        """
        token_array_w_frame = token_array
        atom_level_token_mask = get_atom_level_token_mask(token_array, atom_array)

        # Construct a KDTree for queries to avoid redundant distance calculations
        lig_res_ref_conf_kdtree = {}
        # Ligand and non-standard residues need to use ref to identify frames
        lig_atom_array = atom_array[
            (atom_array.mol_type == "ligand")
            | (~np.isin(atom_array.res_name, list(STD_RESIDUES.keys())))
            | atom_level_token_mask
        ]
        for ref_space_uid in np.unique(lig_atom_array.ref_space_uid):
            # The ref_space_uid is the unique identifier ID for each residue.
            atom_ids = np.where(atom_array.ref_space_uid == ref_space_uid)[0]
            if len(atom_ids) >= 3:
                kdtree = KDTree(ref_pos[atom_ids], metric="euclidean")
            else:
                # Invalid frame
                kdtree = None
            lig_res_ref_conf_kdtree[ref_space_uid] = (kdtree, atom_ids)

        has_frame = []
        for token in token_array_w_frame:
            centre_atom = atom_array[token.centre_atom_index]
            if (
                centre_atom.mol_type != "ligand"
                and centre_atom.res_name in STD_RESIDUES
                and len(token.atom_indices) > 1
            ):
                has_frame, frame_atom_index = Featurizer.get_prot_nuc_frame(
                    token, centre_atom
                )

            else:
                has_frame, frame_atom_index = Featurizer.get_lig_frame(
                    token, centre_atom, lig_res_ref_conf_kdtree, ref_pos, ref_mask
                )

            token.has_frame = has_frame
            token.frame_atom_index = frame_atom_index
        return token_array_w_frame

    def get_token_features(self) -> dict[str, torch.Tensor]:
        """
        Ref: AlphaFold3 SI Chapter 2.8

        Get token features.
        The size of these features is [N_token].

        Returns:
            Dict[str, torch.Tensor]: A dict of token features.
        """
        token_features = {}

        centre_atoms_indices = self.cropped_token_array.get_annotation(
            "centre_atom_index"
        )
        centre_atoms = self.cropped_atom_array[centre_atoms_indices]

        restype = centre_atoms.cano_seq_resname
        restype_onehot = self.restype_onehot_encoded(restype)

        token_features["token_index"] = torch.arange(0, len(self.cropped_token_array))
        token_features["residue_index"] = torch.Tensor(
            centre_atoms.res_id.astype(int)
        ).long()
        token_features["asym_id"] = torch.Tensor(centre_atoms.asym_id_int).long()
        token_features["entity_id"] = torch.Tensor(centre_atoms.entity_id_int).long()
        token_features["sym_id"] = torch.Tensor(centre_atoms.sym_id_int).long()
        token_features["restype"] = restype_onehot

        return token_features

    def get_chain_perm_features(self) -> dict[str, torch.Tensor]:
        """
        The chain permutation use "entity_mol_id", "mol_id" and "mol_atom_index"
        instead of the "entity_id", "asym_id" and "residue_index".

        The shape of these features is [N_atom].

        Returns:
            Dict[str, torch.Tensor]: A dict of chain permutation features.
        """

        chain_perm_features = {}
        chain_perm_features["mol_id"] = torch.Tensor(
            self.cropped_atom_array.mol_id
        ).long()
        chain_perm_features["mol_atom_index"] = torch.Tensor(
            self.cropped_atom_array.mol_atom_index
        ).long()
        chain_perm_features["entity_mol_id"] = torch.Tensor(
            self.cropped_atom_array.entity_mol_id
        ).long()
        return chain_perm_features

    def get_renamed_atom_names(self) -> np.ndarray:
        """
        Rename the atom names of ligands to avioid information leakage.

        Returns:
            np.ndarray: A numpy array of renamed atom names.
        """
        res_starts = get_residue_starts(
            self.cropped_atom_array, add_exclusive_stop=True
        )
        new_atom_names = copy.deepcopy(self.cropped_atom_array.atom_name)
        for start, stop in zip(res_starts[:-1], res_starts[1:]):
            res_mol_type = self.cropped_atom_array.mol_type[start]
            if res_mol_type != "ligand":
                continue

            elem_count = defaultdict(int)
            new_res_atom_names = []
            for elem in self.cropped_atom_array.element[start:stop]:
                elem_count[elem] += 1
                new_res_atom_names.append(f"{elem.upper()}{elem_count[elem]}")
            new_atom_names[start:stop] = new_res_atom_names
        return new_atom_names

    def get_reference_features(self) -> dict[str, torch.Tensor]:
        """
        Ref: AlphaFold3 SI Chapter 2.8

        Get reference features.
        The size of these features is [N_atom].

        Returns:
            Dict[str, torch.Tensor]: a dict of reference features.
        """
        ref_pos = []
        for ref_space_uid in np.unique(self.cropped_atom_array.ref_space_uid):
            res_ref_pos = random_transform(
                self.cropped_atom_array.ref_pos[
                    self.cropped_atom_array.ref_space_uid == ref_space_uid,
                ],
                apply_augmentation=self.ref_pos_augment,
                centralize=True,
            )
            ref_pos.append(res_ref_pos)
        ref_pos = np.concatenate(ref_pos)

        ref_features = {}
        ref_features["ref_pos"] = torch.Tensor(ref_pos)
        ref_features["ref_mask"] = torch.Tensor(self.cropped_atom_array.ref_mask).long()
        ref_features["ref_element"] = Featurizer.elem_onehot_encoded(
            self.cropped_atom_array.element
        ).long()
        ref_features["ref_charge"] = torch.Tensor(
            self.cropped_atom_array.ref_charge
        ).long()

        if self.lig_atom_rename:
            atom_names = self.get_renamed_atom_names()
        else:
            atom_names = self.cropped_atom_array.atom_name

        ref_features["ref_atom_name_chars"] = Featurizer.ref_atom_name_chars_encoded(
            atom_names
        ).long()
        ref_features["ref_space_uid"] = torch.Tensor(
            self.cropped_atom_array.ref_space_uid
        ).long()

        token_array_with_frame = self.get_token_frame(
            token_array=self.cropped_token_array,
            atom_array=self.cropped_atom_array,
            ref_pos=ref_features["ref_pos"],
            ref_mask=ref_features["ref_mask"],
        )
        ref_features["has_frame"] = torch.Tensor(
            token_array_with_frame.get_annotation("has_frame")
        ).long()  # [N_token]
        ref_features["frame_atom_index"] = torch.Tensor(
            token_array_with_frame.get_annotation("frame_atom_index")
        ).long()  # [N_token, 3]
        return ref_features

    def get_bond_features(self) -> dict[str, torch.Tensor]:
        """
        Ref: AlphaFold3 SI Chapter 2.8
        A 2D matrix indicating if there is a bond between any atom in token i and token j,
        restricted to just polymer-ligand and ligand-ligand bonds and bonds less than 2.4 Å during training.
        The size of bond feature is [N_token, N_token].
        Returns:
            Dict[str, torch.Tensor]: A dict of bond features.
        """
        bond_array = self.cropped_atom_array.bonds.as_array()
        bond_atom_i = bond_array[:, 0]
        bond_atom_j = bond_array[:, 1]
        ref_space_uid = self.cropped_atom_array.ref_space_uid
        polymer_mask = np.isin(
            self.cropped_atom_array.mol_type, ["protein", "dna", "rna"]
        )
        std_res_mask = (
            np.isin(self.cropped_atom_array.res_name, list(STD_RESIDUES.keys()))
            & polymer_mask
        )
        unstd_res_mask = ~std_res_mask & polymer_mask
        # the polymer-polymer (std-std, std-unstd, and inter-unstd) bond will not be included in token_bonds.
        std_std_bond_mask = std_res_mask[bond_atom_i] & std_res_mask[bond_atom_j]
        std_unstd_bond_mask = (
            std_res_mask[bond_atom_i] & unstd_res_mask[bond_atom_j]
        ) | (std_res_mask[bond_atom_j] & unstd_res_mask[bond_atom_i])
        inter_unstd_bond_mask = (
            unstd_res_mask[bond_atom_i] & unstd_res_mask[bond_atom_j]
        ) & (ref_space_uid[bond_atom_i] != ref_space_uid[bond_atom_j])

        kept_mask = ~(std_std_bond_mask | std_unstd_bond_mask | inter_unstd_bond_mask)
        if self.include_discont_poly_poly_bonds:
            # include discontinuous polymer-polymer bonds
            res_id_i = self.cropped_atom_array.res_id[bond_atom_i]
            res_id_j = self.cropped_atom_array.res_id[bond_atom_j]
            chain_i = self.cropped_atom_array.chain_id[bond_atom_i]
            chain_j = self.cropped_atom_array.chain_id[bond_atom_j]
            is_discont = (np.abs(res_id_i - res_id_j) > 1) | (chain_i != chain_j)
            kept_mask |= is_discont

        kept_bonds = bond_array[kept_mask]

        # -1 means the atom is not in any token
        atom_idx_to_token_idx = np.zeros(len(self.cropped_atom_array), dtype=int) - 1
        for idx, token in enumerate(self.cropped_token_array.tokens):
            for atom_idx in token.atom_indices:
                atom_idx_to_token_idx[atom_idx] = idx
        assert np.all(atom_idx_to_token_idx >= 0), "Some atoms are not in any token"
        num_tokens = len(self.cropped_token_array)
        token_adj_matrix = np.zeros((num_tokens, num_tokens), dtype=int)
        bond_token_i, bond_atom_j = (
            atom_idx_to_token_idx[kept_bonds[:, 0]],
            atom_idx_to_token_idx[kept_bonds[:, 1]],
        )
        for i, j in zip(bond_token_i, bond_atom_j):
            token_adj_matrix[i, j] = 1
            token_adj_matrix[j, i] = 1
        bond_features = {"token_bonds": torch.Tensor(token_adj_matrix)}
        return bond_features

    def get_extra_features(self) -> dict[str, torch.Tensor]:
        """
        Get other features not listed in AlphaFold3 SI Chapter 2.8 Table 5.
        The size of these features is [N_atom].

        Returns:
            Dict[str, torch.Tensor]: a dict of extra features.
        """
        atom_to_token_idx_dict = {}
        for idx, token in enumerate(self.cropped_token_array.tokens):
            for atom_idx in token.atom_indices:
                atom_to_token_idx_dict[atom_idx] = idx

        # Ensure the order of the atom_to_token_idx is the same as the atom_array
        atom_to_token_idx = [
            atom_to_token_idx_dict[atom_idx]
            for atom_idx in range(len(self.cropped_atom_array))
        ]

        extra_features = {}
        extra_features["atom_to_token_idx"] = torch.Tensor(atom_to_token_idx).long()
        extra_features["atom_to_tokatom_idx"] = torch.Tensor(
            self.cropped_atom_array.tokatom_idx
        ).long()

        extra_features["is_protein"] = torch.Tensor(
            self.cropped_atom_array.is_protein
        ).long()
        extra_features["is_ligand"] = torch.Tensor(
            self.cropped_atom_array.is_ligand
        ).long()
        extra_features["is_dna"] = torch.Tensor(self.cropped_atom_array.is_dna).long()
        extra_features["is_rna"] = torch.Tensor(self.cropped_atom_array.is_rna).long()
        if "resolution" in self.cropped_atom_array._annot:
            extra_features["resolution"] = torch.Tensor(
                [self.cropped_atom_array.resolution[0]]
            )
        else:
            extra_features["resolution"] = torch.Tensor([-1])
        return extra_features

    @staticmethod
    def get_lig_pocket_mask(
        atom_array: AtomArray, lig_label_asym_id: Union[str, list]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Ref: AlphaFold3 Chapter Methods.Metrics

        the pocket is defined as all heavy atoms within 10 Å of any heavy atom of the ligand,
        restricted to the primary polymer chain for the ligand or modified residue being scored,
        and further restricted to only backbone atoms for proteins. The primary polymer chain is defined variously:
        for PoseBusters it is the protein chain with the most atoms within 10 Å of the ligand,
        for bonded ligand scores it is the bonded polymer chain and for modified residues it
        is the chain that the residue is contained in (minus that residue).

        Args:
            atom_array (AtomArray): atoms in the complex.
            lig_label_asym_id (Union[str, List]): The label_asym_id of the ligand of interest.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: A tuple of ligand pocket mask and pocket mask.
        """

        if isinstance(lig_label_asym_id, str):
            lig_label_asym_ids = [lig_label_asym_id]
        else:
            lig_label_asym_ids = list(lig_label_asym_id)

        # Get backbone mask
        prot_backbone = (
            atom_array.is_protein & np.isin(atom_array.atom_name, ["C", "N", "CA"])
        ).astype(bool)

        kdtree = KDTree(atom_array.coord)

        ligand_mask_list = []
        pocket_mask_list = []
        for lig_label_asym_id in lig_label_asym_ids:
            assert np.isin(
                lig_label_asym_id, atom_array.label_asym_id
            ), f"{lig_label_asym_id} is not in the label_asym_id of the cropped atom array."

            ligand_mask = atom_array.label_asym_id == lig_label_asym_id
            lig_pos = atom_array.coord[ligand_mask & atom_array.is_resolved]

            # Get atoms in 10 Angstrom radius
            near_atom_indices = np.unique(
                np.concatenate(kdtree.query_radius(lig_pos, 10.0))
            )
            near_atoms = [
                (
                    True
                    if ((i in near_atom_indices) and atom_array.is_resolved[i])
                    else False
                )
                for i in range(len(atom_array))
            ]

            # Get primary chain (protein backone in 10 Angstrom radius)
            primary_chain_candidates = near_atoms & prot_backbone
            primary_chain_candidates_atoms = atom_array[primary_chain_candidates]

            max_atom = 0
            primary_chain_asym_id_int = None
            for asym_id_int in np.unique(primary_chain_candidates_atoms.asym_id_int):
                n_atoms = np.sum(
                    primary_chain_candidates_atoms.asym_id_int == asym_id_int
                )
                if n_atoms > max_atom:
                    max_atom = n_atoms
                    primary_chain_asym_id_int = asym_id_int
            assert (
                primary_chain_asym_id_int is not None
            ), f"No primary chain found for ligand ({lig_label_asym_id=})."

            pocket_mask = primary_chain_candidates & (
                atom_array.asym_id_int == primary_chain_asym_id_int
            )
            ligand_mask_list.append(ligand_mask)
            pocket_mask_list.append(pocket_mask)

        ligand_mask_by_pockets = torch.Tensor(
            np.array(ligand_mask_list).astype(int)
        ).long()
        pocket_mask_by_pockets = torch.Tensor(
            np.array(pocket_mask_list).astype(int)
        ).long()
        return ligand_mask_by_pockets, pocket_mask_by_pockets

    def get_mask_features(self) -> dict[str, torch.Tensor]:
        """
        Generate mask features for the cropped atom array.

        Returns:
            Dict[str, torch.Tensor]: A dictionary containing various mask features.
        """
        mask_features = {}

        mask_features["pae_rep_atom_mask"] = torch.Tensor(
            self.cropped_atom_array.centre_atom_mask
        ).long()

        mask_features["plddt_m_rep_atom_mask"] = torch.Tensor(
            self.cropped_atom_array.plddt_m_rep_atom_mask
        ).long()  # [N_atom]

        mask_features["distogram_rep_atom_mask"] = torch.Tensor(
            self.cropped_atom_array.distogram_rep_atom_mask
        ).long()  # [N_atom]

        mask_features["modified_res_mask"] = torch.Tensor(
            self.cropped_atom_array.modified_res_mask
        ).long()

        lig_polymer_bonds = get_ligand_polymer_bond_mask(self.cropped_atom_array)
        num_atoms = len(self.cropped_atom_array)
        bond_mask_mat = np.zeros((num_atoms, num_atoms))
        for i, j, _ in lig_polymer_bonds:
            bond_mask_mat[i, j] = 1
            bond_mask_mat[j, i] = 1
        mask_features["bond_mask"] = torch.Tensor(
            bond_mask_mat
        ).long()  # [N_atom, N_atom]
        return mask_features

    def get_all_input_features(self):
        """
        Get input features from cropped data.

        Returns:
            Dict[str, torch.Tensor]: a dict of features.
        """
        features = {}
        token_features = self.get_token_features()
        features.update(token_features)

        bond_features = self.get_bond_features()
        features.update(bond_features)

        reference_features = self.get_reference_features()
        features.update(reference_features)

        extra_features = self.get_extra_features()
        features.update(extra_features)

        chain_perm_features = self.get_chain_perm_features()
        features.update(chain_perm_features)

        mask_features = self.get_mask_features()
        features.update(mask_features)
        return features

    def get_labels(self) -> dict[str, torch.Tensor]:
        """
        Get the input labels required for the training phase.

        Returns:
            Dict[str, torch.Tensor]: a dict of labels.
        """

        labels = {}

        labels["coordinate"] = torch.Tensor(
            self.cropped_atom_array.coord
        )  # [N_atom, 3]

        labels["coordinate_mask"] = torch.Tensor(
            self.cropped_atom_array.is_resolved.astype(int)
        ).long()  # [N_atom]
        return labels

    def get_atom_permutation_list(
        self,
    ) -> list[list[int]]:
        """
        Generate info of permutations.

        Returns:
            List[List[int]]: a list of atom permutations.
        """
        atom_perm_list = []
        for i in self.cropped_atom_array.res_perm:
            # Decode list[str] -> list[list[int]]
            atom_perm_list.append([int(j) for j in i.split("_")])

        # Atoms connected to different residue are fixed.
        # Bonds array: [[atom_idx_i, atom_idx_j, bond_type]]
        idx_i = self.cropped_atom_array.bonds._bonds[:, 0]
        idx_j = self.cropped_atom_array.bonds._bonds[:, 1]
        diff_mask = (
            self.cropped_atom_array.ref_space_uid[idx_i]
            != self.cropped_atom_array.ref_space_uid[idx_j]
        )
        inter_residue_bonds = self.cropped_atom_array.bonds._bonds[diff_mask]
        fixed_atom_mask = np.isin(
            np.arange(len(self.cropped_atom_array)),
            np.unique(inter_residue_bonds[:, :2]),
        )

        # Get fixed atom permutation for each residue.
        fixed_atom_perm_list = []
        res_starts = get_residue_starts(
            self.cropped_atom_array, add_exclusive_stop=True
        )
        for r_start, r_stop in zip(res_starts[:-1], res_starts[1:]):
            atom_res_perm = np.array(
                atom_perm_list[r_start:r_stop]
            )  # [N_res_atoms, N_res_perm]
            res_fixed_atom_mask = fixed_atom_mask[r_start:r_stop]

            if np.sum(res_fixed_atom_mask) == 0:
                # If all atoms in the residue are not fixed, e.g. ions
                fixed_atom_perm_list.extend(atom_res_perm.tolist())
                continue

            # Create a [N_res_atoms, N_res_perm] template of indices
            n_res_atoms, n_perm = atom_res_perm.shape
            indices_template = (
                atom_res_perm[:, 0].reshape(n_res_atoms, 1).repeat(n_perm, axis=1)
            )

            # Identify the column where the positions of the fixed atoms remain unchanged
            fixed_atom_perm = atom_res_perm[
                res_fixed_atom_mask
            ]  # [N_fixed_res_atoms, N_res_perm]
            fixed_indices_template = indices_template[
                res_fixed_atom_mask
            ]  # [N_fixed_res_atoms, N_res_perm]
            unchanged_columns_mask = np.all(
                fixed_atom_perm == fixed_indices_template, axis=0
            )

            # Remove the columns related to the position changes of fixed atoms.
            fixed_atom_res_perm = atom_res_perm[:, unchanged_columns_mask]
            fixed_atom_perm_list.extend(fixed_atom_res_perm.tolist())
        return fixed_atom_perm_list

    @staticmethod
    def get_gt_full_complex_features(
        atom_array: AtomArray,
        cropped_atom_array: AtomArray = None,
        get_cropped_asym_only: bool = True,
    ) -> dict[str, torch.Tensor]:
        """Get full ground truth complex features.
        It is used for multi-chain permutation alignment.

        Args:
            atom_array (AtomArray): all atoms in the complex.
            cropped_atom_array (AtomArray, optional): cropped atoms. Defaults to None.
            get_cropped_asym_only (bool, optional): Defaults to True.
                - If true, a chain is returned only if its asym_id (mol_id) appears in the
                cropped_atom_array. It should be a favored setting for the spatial cropping.
                - If false, a chain is returned if its entity_id (entity_mol_id) appears in
                the cropped_atom_array.

        Returns:
            Dict[str, torch.Tensor]: a dictionary containing
                coordinate, coordinate_mask, etc.
        """
        gt_features = {}

        if cropped_atom_array is not None:
            # Get the cropped part of gt entities
            entity_atom_set = set(
                zip(
                    cropped_atom_array.entity_mol_id,
                    cropped_atom_array.mol_atom_index,
                )
            )
            mask = [
                (entity, atom) in entity_atom_set
                for (entity, atom) in zip(
                    atom_array.entity_mol_id, atom_array.mol_atom_index
                )
            ]

            if get_cropped_asym_only:
                # Restrict to asym chains appeared in cropped_atom_array
                asyms = np.unique(cropped_atom_array.mol_id)
                mask = mask * np.isin(atom_array.mol_id, asyms)
            atom_array = atom_array[mask]

        gt_features["coordinate"] = torch.Tensor(atom_array.coord)
        gt_features["coordinate_mask"] = torch.Tensor(atom_array.is_resolved).long()
        gt_features["entity_mol_id"] = torch.Tensor(atom_array.entity_mol_id).long()
        gt_features["mol_id"] = torch.Tensor(atom_array.mol_id).long()
        gt_features["mol_atom_index"] = torch.Tensor(atom_array.mol_atom_index).long()
        gt_features["pae_rep_atom_mask"] = torch.Tensor(
            atom_array.centre_atom_mask
        ).long()
        return gt_features, atom_array



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Optional

import torch

from protenix.model.utils import centre_random_augmentation


class TrainingNoiseSampler:
    """
    Sample the noise-level of training samples.

    Args:
        p_mean (float, optional): gaussian mean. Defaults to -1.2.
        p_std (float, optional): gaussian std. Defaults to 1.5.
        sigma_data (float, optional): scale. Defaults to 16.0, but this is 1.0 in EDM.
    """

    def __init__(
        self,
        p_mean: float = -1.2,
        p_std: float = 1.5,
        sigma_data: float = 16.0,  # NOTE: in EDM, this is 1.0
    ) -> None:
        self.sigma_data = sigma_data
        self.p_mean = p_mean
        self.p_std = p_std
        print(f"train scheduler {self.sigma_data}")

    def __call__(
        self, size: torch.Size, device: torch.device = torch.device("cpu")
    ) -> torch.Tensor:
        """Sampling

        Args:
            size (torch.Size): the target size
            device (torch.device, optional): target device. Defaults to torch.device("cpu").

        Returns:
            torch.Tensor: sampled noise-level
        """
        rnd_normal = torch.randn(size=size, device=device)
        noise_level = (rnd_normal * self.p_std + self.p_mean).exp() * self.sigma_data
        return noise_level


class InferenceNoiseScheduler:
    """
    Scheduler for noise-level (time steps).

    Args:
        s_max (float, optional): maximal noise level. Defaults to 160.0.
        s_min (float, optional): minimal noise level. Defaults to 4e-4.
        rho (float, optional): the exponent numerical part. Defaults to 7.
        sigma_data (float, optional): scale. Defaults to 16.0, but this is 1.0 in EDM.
    """

    def __init__(
        self,
        s_max: float = 160.0,
        s_min: float = 4e-4,
        rho: float = 7,
        sigma_data: float = 16.0,  # NOTE: in EDM, this is 1.0
    ) -> None:
        self.sigma_data = sigma_data
        self.s_max = s_max
        self.s_min = s_min
        self.rho = rho
        print(f"inference scheduler {self.sigma_data}")

    def __call__(
        self,
        N_step: int = 200,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.float32,
    ) -> torch.Tensor:
        """Schedule the noise-level (time steps). No sampling is performed.

        Args:
            N_step (int, optional): number of time steps. Defaults to 200.
            device (torch.device, optional): target device. Defaults to torch.device("cpu").
            dtype (torch.dtype, optional): target dtype. Defaults to torch.float32.

        Returns:
            torch.Tensor: noise-level (time_steps)
                [N_step+1]
        """
        step_size = 1 / N_step
        step_indices = torch.arange(N_step + 1, device=device, dtype=dtype)
        t_step_list = (
            self.sigma_data
            * (
                self.s_max ** (1 / self.rho)
                + step_indices
                * step_size
                * (self.s_min ** (1 / self.rho) - self.s_max ** (1 / self.rho))
            )
            ** self.rho
        )
        # replace the last time step by 0
        t_step_list[..., -1] = 0  # t_N = 0

        return t_step_list


def sample_diffusion(
    denoise_net: Callable,
    input_feature_dict: dict[str, Any],
    s_inputs: torch.Tensor,
    s_trunk: torch.Tensor,
    z_trunk: torch.Tensor,
    pair_z: torch.Tensor,
    p_lm: torch.Tensor,
    c_l: torch.Tensor,
    noise_schedule: torch.Tensor,
    N_sample: int = 1,
    gamma0: float = 0.8,
    gamma_min: float = 1.0,
    noise_scale_lambda: float = 1.003,
    step_scale_eta: float = 1.5,
    diffusion_chunk_size: Optional[int] = None,
    inplace_safe: bool = False,
    attn_chunk_size: Optional[int] = None,
    enable_efficient_fusion: bool = False,
) -> torch.Tensor:
    """Implements Algorithm 18 in AF3.
    It performances denoising steps from time 0 to time T.
    The time steps (=noise levels) are given by noise_schedule.

    Args:
        denoise_net (Callable): the network that performs the denoising step.
        input_feature_dict (dict[str, Any]): input meta feature dict
        s_inputs (torch.Tensor): single embedding from InputFeatureEmbedder
            [..., N_tokens, c_s_inputs]
        s_trunk (torch.Tensor): single feature embedding from PairFormer (Alg17)
            [..., N_tokens, c_s]
        z_trunk (torch.Tensor): pair feature embedding from PairFormer (Alg17)
            [..., N_tokens, N_tokens, c_z]
        pair_z (torch.Tensor): pair feature embedding from InputFeatureEmbedder
            [..., N_tokens, N_tokens, c_z_inputs]
        p_lm (torch.Tensor): MSA embedding
            [..., N_tokens, c_p_lm]
        c_l (torch.Tensor): ligand embedding
            [..., N_tokens, c_c_l]
        noise_schedule (torch.Tensor): noise-level schedule (which is also the time steps) since sigma=t.
            [N_iterations]
        N_sample (int): number of generated samples
        gamma0 (float): params in Alg.18.
        gamma_min (float): params in Alg.18.
        noise_scale_lambda (float): params in Alg.18.
        step_scale_eta (float): params in Alg.18.
        diffusion_chunk_size (Optional[int]): Chunk size for diffusion operation. Defaults to None.
        inplace_safe (bool): Whether to use inplace operations safely. Defaults to False.
        attn_chunk_size (Optional[int]): Chunk size for attention operation. Defaults to None.
        enable_efficient_fusion (bool): Whether to enable efficient fusion. Defaults to False.

    Returns:
        torch.Tensor: the denoised coordinates of x in inference stage
            [..., N_sample, N_atom, 3]
    """
    N_atom = input_feature_dict["atom_to_token_idx"].size(-1)
    batch_shape = s_inputs.shape[:-2]
    device = s_inputs.device
    dtype = s_inputs.dtype

    def _chunk_sample_diffusion(chunk_n_sample, inplace_safe):
        # init noise
        # [..., N_sample, N_atom, 3]
        x_l = noise_schedule[0] * torch.randn(
            size=(*batch_shape, chunk_n_sample, N_atom, 3), device=device, dtype=dtype
        )  # NOTE: set seed in distributed training

        for _, (c_tau_last, c_tau) in enumerate(
            zip(noise_schedule[:-1], noise_schedule[1:])
        ):
            # [..., N_sample, N_atom, 3]
            x_l = (
                centre_random_augmentation(x_input_coords=x_l, N_sample=1)
                .squeeze(dim=-3)
                .to(dtype)
            )

            # Denoise with a predictor-corrector sampler
            # 1. Add noise to move x_{c_tau_last} to x_{t_hat}
            gamma = float(gamma0) if c_tau > gamma_min else 0
            t_hat = c_tau_last * (gamma + 1)

            delta_noise_level = torch.sqrt(t_hat**2 - c_tau_last**2)
            x_noisy = x_l + noise_scale_lambda * delta_noise_level * torch.randn(
                size=x_l.shape, device=device, dtype=dtype
            )

            # 2. Denoise from x_{t_hat} to x_{c_tau}
            # Euler step only
            t_hat = (
                t_hat.reshape((1,) * (len(batch_shape) + 1))
                .expand(*batch_shape, chunk_n_sample)
                .to(dtype)
            )

            x_denoised = denoise_net(
                x_noisy=x_noisy,
                t_hat_noise_level=t_hat,
                input_feature_dict=input_feature_dict,
                s_inputs=s_inputs,
                s_trunk=s_trunk,
                z_trunk=z_trunk,
                pair_z=pair_z,
                p_lm=p_lm,
                c_l=c_l,
                chunk_size=attn_chunk_size,
                inplace_safe=inplace_safe,
                enable_efficient_fusion=enable_efficient_fusion,
            )

            delta = (x_noisy - x_denoised) / t_hat[
                ..., None, None
            ]  # Line 9 of AF3 uses 'x_l_hat' instead, which we believe  is a typo.
            dt = c_tau - t_hat
            x_l = x_noisy + step_scale_eta * dt[..., None, None] * delta

        return x_l

    if diffusion_chunk_size is None:
        x_l = _chunk_sample_diffusion(N_sample, inplace_safe=inplace_safe)
    else:
        x_l = []
        no_chunks = N_sample // diffusion_chunk_size + (
            N_sample % diffusion_chunk_size != 0
        )
        for i in range(no_chunks):
            chunk_n_sample = (
                diffusion_chunk_size
                if i < no_chunks - 1
                else N_sample - i * diffusion_chunk_size
            )
            chunk_x_l = _chunk_sample_diffusion(
                chunk_n_sample, inplace_safe=inplace_safe
            )
            x_l.append(chunk_x_l)
        x_l = torch.cat(x_l, -3)  # [..., N_sample, N_atom, 3]
    return x_l


def sample_diffusion_training(
    noise_sampler: TrainingNoiseSampler,
    denoise_net: Callable,
    label_dict: dict[str, Any],
    input_feature_dict: dict[str, Any],
    s_inputs: torch.Tensor,
    s_trunk: torch.Tensor,
    z_trunk: torch.Tensor,
    pair_z: torch.Tensor,
    p_lm: torch.Tensor,
    c_l: torch.Tensor,
    N_sample: int = 1,
    diffusion_chunk_size: Optional[int] = None,
    use_conditioning: bool = True,
    enable_efficient_fusion: bool = False,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Implements diffusion training as described in AF3 Appendix at page 23.
    It performances denoising steps from time 0 to time T.
    The time steps (=noise levels) are given by noise_schedule.

    Args:
        noise_sampler (TrainingNoiseSampler): sampler for training noise-level.
        denoise_net (Callable): the network that performs the denoising step.
        label_dict (dict[str, Any]) : a dictionary containing the followings.
            "coordinate": the ground-truth coordinates
                [..., N_atom, 3]
            "coordinate_mask": whether true coordinates exist.
                [..., N_atom]
        input_feature_dict (dict[str, Any]): input meta feature dict
        s_inputs (torch.Tensor): single embedding from InputFeatureEmbedder
            [..., N_tokens, c_s_inputs]
        s_trunk (torch.Tensor): single feature embedding from PairFormer (Alg17)
            [..., N_tokens, c_s]
        z_trunk (torch.Tensor): pair feature embedding from PairFormer (Alg17)
            [..., N_tokens, N_tokens, c_z]
        pair_z (torch.Tensor): pair feature embedding from InputFeatureEmbedder
            [..., N_tokens, N_tokens, c_z_inputs]
        p_lm (torch.Tensor): MSA embedding
            [..., N_tokens, c_p_lm]
        c_l (torch.Tensor): ligand embedding
            [..., N_tokens, c_c_l]
        N_sample (int): number of training samples
        diffusion_chunk_size (Optional[int]): Chunk size for diffusion operation. Defaults to None.
        use_conditioning (bool): Whether to use conditioning. Defaults to True.
        enable_efficient_fusion (bool): Whether to enable efficient fusion. Defaults to False.

    Returns:
        tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
            x_gt_augment: the augmented ground-truth coordinates [..., N_sample, N_atom, 3]
            x_denoised: the denoised coordinates [..., N_sample, N_atom, 3]
            sigma: the sampled noise-level [..., N_sample]
    """
    batch_size_shape = label_dict["coordinate"].shape[:-2]
    device = label_dict["coordinate"].device
    dtype = label_dict["coordinate"].dtype
    # Areate N_sample versions of the input structure by randomly rotating and translating
    x_gt_augment = centre_random_augmentation(
        x_input_coords=label_dict["coordinate"],
        N_sample=N_sample,
        mask=label_dict["coordinate_mask"],
    ).to(
        dtype
    )  # [..., N_sample, N_atom, 3]

    # Add independent noise to each structure
    # sigma: independent noise-level [..., N_sample]
    sigma = noise_sampler(size=(*batch_size_shape, N_sample), device=device).to(dtype)
    # noise: [..., N_sample, N_atom, 3]
    noise = torch.randn_like(x_gt_augment, dtype=dtype) * sigma[..., None, None]

    # Get denoising outputs [..., N_sample, N_atom, 3]
    if diffusion_chunk_size is None:
        x_denoised = denoise_net(
            x_noisy=x_gt_augment + noise,
            t_hat_noise_level=sigma,
            input_feature_dict=input_feature_dict,
            s_inputs=s_inputs,
            s_trunk=s_trunk,
            z_trunk=z_trunk,
            pair_z=pair_z,
            p_lm=p_lm,
            c_l=c_l,
            use_conditioning=use_conditioning,
            enable_efficient_fusion=enable_efficient_fusion,
        )
    else:
        x_denoised = []
        no_chunks = N_sample // diffusion_chunk_size + (
            N_sample % diffusion_chunk_size != 0
        )
        for i in range(no_chunks):
            x_noisy_i = (x_gt_augment + noise)[
                ..., i * diffusion_chunk_size : (i + 1) * diffusion_chunk_size, :, :
            ]
            t_hat_noise_level_i = sigma[
                ..., i * diffusion_chunk_size : (i + 1) * diffusion_chunk_size
            ]
            x_denoised_i = denoise_net(
                x_noisy=x_noisy_i,
                t_hat_noise_level=t_hat_noise_level_i,
                input_feature_dict=input_feature_dict,
                s_inputs=s_inputs,
                s_trunk=s_trunk,
                z_trunk=z_trunk,
                pair_z=pair_z,
                p_lm=p_lm,
                c_l=c_l,
                use_conditioning=use_conditioning,
                enable_efficient_fusion=enable_efficient_fusion,
            )
            x_denoised.append(x_denoised_i)
        x_denoised = torch.cat(x_denoised, dim=-3)

    return x_gt_augment, x_denoised, sigma



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from scipy.spatial.transform import Rotation


def angle_3p(a, b, c):
    """
    Calculate the angle between three points in a 2D space.

    Args:
        a (list or array-like): The coordinates of the first point.
        b (list or array-like): The coordinates of the second point.
        c (list or array-like): The coordinates of the third point.

    Returns:
        float: The angle in degrees (0, 180) between the vectors
               from point a to point b and point b to point c.
    """
    a = np.asarray(a)
    b = np.asarray(b)
    c = np.asarray(c)

    ab = b - a
    bc = c - b

    dot_product = np.dot(ab, bc)

    norm_ab = np.linalg.norm(ab)
    norm_bc = np.linalg.norm(bc)

    cos_theta = np.clip(dot_product / (norm_ab * norm_bc + 1e-4), -1, 1)
    theta_radians = np.arccos(cos_theta)
    theta_degrees = np.degrees(theta_radians)
    return theta_degrees


def random_transform(
    points, max_translation=1.0, apply_augmentation=False, centralize=True
) -> np.ndarray:
    """
    Randomly transform a set of 3D points.

    Args:
        points (numpy.ndarray): The points to be transformed, shape=(N, 3)
        max_translation (float): The maximum translation value. Default is 1.0.
        apply_augmentation (bool): Whether to apply random rotation/translation on ref_pos

    Returns:
        numpy.ndarray: The transformed points.
    """
    if centralize:
        points = points - points.mean(axis=0)
    if not apply_augmentation:
        return points
    translation = np.random.uniform(-max_translation, max_translation, size=3)
    R = Rotation.random().as_matrix()
    transformed_points = np.dot(points + translation, R.T)
    return transformed_points



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn.functional as F

from protenix.model.utils import batched_gather


def expressCoordinatesInFrame(
    coordinate: torch.Tensor, frames: torch.Tensor, eps: float = 1e-8
) -> torch.Tensor:
    """Algorithm 29 Express coordinate in frame

    Args:
        coordinate (torch.Tensor): the input coordinate
            [..., N_atom, 3]
        frames (torch.Tensor): the input frames
            [..., N_frame, 3, 3]
        eps (float): Small epsilon value

    Returns:
        torch.Tensor: the transformed coordinate projected onto frame basis
            [..., N_frame, N_atom, 3]
    """
    # Extract frame atoms
    a, b, c = torch.unbind(frames, dim=-2)  # a, b, c shape: [..., N_frame, 3]
    w1 = F.normalize(a - b, dim=-1, eps=eps)
    w2 = F.normalize(c - b, dim=-1, eps=eps)
    # Build orthonormal basis
    e1 = F.normalize(w1 + w2, dim=-1, eps=eps)
    e2 = F.normalize(w2 - w1, dim=-1, eps=eps)
    e3 = torch.cross(e1, e2, dim=-1)  # [..., N_frame, 3]
    # Project onto frame basis
    d = coordinate[..., None, :, :] - b[..., None, :]  # [..., N_frame, N_atom, 3]
    x_transformed = torch.cat(
        [
            torch.sum(d * e1[..., None, :], dim=-1, keepdim=True),
            torch.sum(d * e2[..., None, :], dim=-1, keepdim=True),
            torch.sum(d * e3[..., None, :], dim=-1, keepdim=True),
        ],
        dim=-1,
    )  # [..., N_frame, N_atom, 3]
    return x_transformed


def gather_frame_atom_by_indices(
    coordinate: torch.Tensor, frame_atom_index: torch.Tensor, dim: int = -2
) -> torch.Tensor:
    """construct frames from coordinate

    Args:
        coordinate (torch.Tensor):  the input coordinate
            [..., N_atom, 3]
        frame_atom_index (torch.Tensor): indices of three atoms in each frame
            [..., N_frame, 3] or [N_frame, 3]
        dim (int): along which dimension to select the frame atoms
    Returns:
        torch.Tensor: the constructed frames
            [..., N_frame, 3[three atom], 3[three coordinate]]
    """
    if len(frame_atom_index.shape) == 2:
        # the navie case
        x1 = torch.index_select(
            coordinate, dim=dim, index=frame_atom_index[:, 0]
        )  # [..., N_frame, 3]
        x2 = torch.index_select(
            coordinate, dim=dim, index=frame_atom_index[:, 1]
        )  # [..., N_frame, 3]
        x3 = torch.index_select(
            coordinate, dim=dim, index=frame_atom_index[:, 2]
        )  # [..., N_frame, 3]
        return torch.stack([x1, x2, x3], dim=dim)
    else:
        assert (
            frame_atom_index.shape[:dim] == coordinate.shape[:dim]
        ), "batch size dims should match"

    x1 = batched_gather(
        data=coordinate,
        inds=frame_atom_index[..., 0],
        dim=dim,
        no_batch_dims=len(coordinate.shape[:dim]),
    )  # [..., N_frame, 3]
    x2 = batched_gather(
        data=coordinate,
        inds=frame_atom_index[..., 1],
        dim=dim,
        no_batch_dims=len(coordinate.shape[:dim]),
    )  # [..., N_frame, 3]
    x3 = batched_gather(
        data=coordinate,
        inds=frame_atom_index[..., 2],
        dim=dim,
        no_batch_dims=len(coordinate.shape[:dim]),
    )  # [..., N_frame, 3]
    return torch.stack([x1, x2, x3], dim=dim)



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import warnings

import torch
from torch.optim.lr_scheduler import ConstantLR, LRScheduler


class CosineAnnealingWithWarmup(LRScheduler):
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        warmup_steps: int,
        decay_steps: int,
        lr: float,
        min_lr: float,
        last_epoch: int = -1,
    ):
        self.warmup_steps = warmup_steps
        self.decay_steps = decay_steps
        self.lr = lr
        self.min_lr = min_lr
        super().__init__(optimizer, last_epoch)

    def _get_step_lr(self, step):
        if step <= self.warmup_steps:
            return (step + 1) / (self.warmup_steps + 1) * self.lr
        elif step >= self.decay_steps:
            return self.min_lr
        else:
            decay_ratio = (step - self.warmup_steps) / (
                self.decay_steps - self.warmup_steps
            )
            assert 0 <= decay_ratio <= 1
            coff = 0.5 * (1.0 + math.cos(math.pi * decay_ratio))
            return self.min_lr + coff * (self.lr - self.min_lr)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, "
                "please use `get_last_lr()`.",
                UserWarning,
            )
        return [
            self._get_step_lr(self.last_epoch) for group in self.optimizer.param_groups
        ]

    def _get_closed_form_lr(self):
        return [self._get_step_lr(self.last_epoch) for base_lr in self.base_lrs]


# The Alphafold3 Learning Rate Scheduler As in 5.4
class AlphaFold3LRScheduler(LRScheduler):
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        last_epoch: int = -1,
        warmup_steps: int = 1000,
        lr: float = 1.8e-3,
        decay_every_n_steps: int = 50000,
        decay_factor: float = 0.95,
    ) -> None:
        self.warmup_steps = warmup_steps
        self.decay_steps = decay_every_n_steps
        self.lr = lr
        self.decay_factor = decay_factor
        super(AlphaFold3LRScheduler, self).__init__(
            optimizer=optimizer, last_epoch=last_epoch
        )

    def _get_step_lr(self, step):
        if step <= self.warmup_steps:
            lr = step / self.warmup_steps * self.lr
        else:
            decay_count = step // self.decay_steps
            lr = self.lr * (self.decay_factor**decay_count)
        return lr

    def get_lr(self) -> list[float]:
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, "
                "please use `get_last_lr()`.",
                UserWarning,
            )
        return [
            self._get_step_lr(self.last_epoch) for group in self.optimizer.param_groups
        ]


class ConstantLRScheduler(ConstantLR):
    def __init__(self, optimizer, lr, last_epoch=-1):
        self.lr = lr
        super(ConstantLRScheduler, self).__init__(
            optimizer, factor=1.0, last_epoch=last_epoch
        )

    def _get_step_lr(self, step):
        return self.lr


class FinetuneLRScheduler(LRScheduler):
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        base_lr_config,
        finetune_lr_config,
        last_epoch: int = -1,
    ) -> None:

        self.lr_scheduler = get_lr_scheduler(base_lr_config, optimizer)
        self.finetune_lr_scheduler = get_lr_scheduler(finetune_lr_config, optimizer)
        super(FinetuneLRScheduler, self).__init__(
            optimizer=optimizer, last_epoch=last_epoch
        )

    def _get_step_lr(self, step):
        lr = self.lr_scheduler._get_step_lr(step)
        ft_lr = self.finetune_lr_scheduler._get_step_lr(step)
        # this order is same to the order in optimizer
        return ft_lr, lr

    def get_lr(self) -> list[float]:
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, "
                "please use `get_last_lr()`.",
                UserWarning,
            )
        return self._get_step_lr(self.last_epoch)


def get_lr_scheduler(
    configs, optimizer: torch.optim.Optimizer, **kwargs
) -> torch.optim.lr_scheduler.LRScheduler:
    """
    Get the learning rate scheduler based on the configuration.

    Args:
        configs: Configuration object containing scheduler settings.
        optimizer (torch.optim.Optimizer): The optimizer to which the scheduler will be attached.
        **kwargs: Additional keyword arguments to be passed to the scheduler.

    Returns:
        torch.optim.lr_scheduler.LRScheduler: The learning rate scheduler.

    Raises:
        ValueError: If the specified learning rate scheduler is invalid.
    """
    if configs.lr_scheduler == "af3":
        lr_scheduler = AlphaFold3LRScheduler(
            optimizer, **configs.af3_lr_scheduler, **kwargs
        )
    elif configs.lr_scheduler == "cosine_annealing":
        lr_scheduler = CosineAnnealingWithWarmup(
            optimizer,
            configs.warmup_steps,
            configs.max_steps,
            configs.lr,
            configs.lr * configs.min_lr_ratio,
            **kwargs,
        )
    elif configs.lr_scheduler == "constant":
        lr_scheduler = ConstantLRScheduler(
            optimizer,
            lr=configs.lr,
            **kwargs,
        )
    else:
        raise ValueError(f"Invalid lr scheduler: [{configs.lr_scheduler}]")
    return lr_scheduler
